from ..base_executor import BaseExecutor
from ...insight import InsightState


class CancelAllOppositeSidetExecutor(BaseExecutor):
    """
    ### Executor for Cancel All Opposite Side
    This executor is used to cancel all opposite side Insights generated for the insight symbol.
    This is done by setting the market changed flag to True for all opposite side insights and closing or cancelling the trades insights.

    :param strategy (BaseStrategy): The strategy instance

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.NEW, "1.0")

    def run(self, insight):
        try:
            # Check if we have any current positions for the symbol
            holding = self.STRATEGY.positions.get(insight.symbol)
            if holding is None:
                return self.returnResults(True, True, "No current positions for symbol.")

            if holding["side"] != insight.side:
                # Set the market changed flag to True for all opposite side insights
                numberOfInsightsAffected = 0
                numberOfInsightsFailed = 0

                for id, otherInsight in self.STRATEGY.insights.items():
                    if otherInsight.symbol == insight.symbol and otherInsight.side != insight.side:
                        try:
                            if self.STRATEGY.insights[otherInsight.INSIGHT_ID].update_market_changed(marketChanged=True, shouldCloseOrCancel=True):
                                numberOfInsightsAffected += 1
                        except Exception as e:
                            numberOfInsightsFailed += 1
                            continue

                return self.returnResults(True, True, f"Updated {numberOfInsightsAffected} insights. Failed to update {numberOfInsightsFailed} insights.")
            return self.returnResults(True, True)
        except Exception as e:
            return self.returnResults(False, False, f"Error rejecting expired insight: {e}")




from dash import html

from OlympusTrader.insight.insight import IInsight, InsightState


def dashboardCard(title: str, value: float, icon: str = ""):
    return html.Div([
        html.Div([
            html.H3(title, className="text-xl font-semibold mb-2 text-accent"),
            html.I(None, className=icon)
        ]),
        html.P(value, className="text-2xl font-bold text-white")
    ], className="bg-primary-foreground p-6 rounded-lg shadow-lg border border-accent")

def dashboardCardInsights (title: str, insights: dict[str, IInsight]):
    insightsByStatus = {}
    for insight in insights.values():
        if insight['state'] not in insightsByStatus:
            insightsByStatus[insight['state']] = 1
        else:
            insightsByStatus[insight['state']] += 1

    displayInsights_byStatus = [ html.Div([
                html.P(status, className="text-sm text-accent"),
                html.P(count, className="text-2xl font-bold text-white")
            ], className="flex items-center justify-between") for status, count in insightsByStatus.items()]

    # numberOfFilledInsights = len([insight for insight in insights.values(
    # ) if insight['state'] == str(InsightState.FILLED)])
    return html.Div([
        html.Div([
            html.H3(title, className="text-xl font-semibold mb-2 text-accent"),
        ]),

        html.Div([
            html.P("Total", className="text-sm text-accent"),
            html.P(len(insights), className="text-2xl font-bold text-white")
        ], className="flex items-center justify-between"),
        *displayInsights_byStatus
        
        # html.P(value, className="text-2xl font-bold text-white")
    ], className="bg-primary-foreground p-6 rounded-lg shadow-lg border border-accent")


def dashboardCardMode(title: str, mode: str, accountType: str, connected: bool):
    return html.Div([
        html.Div([
            html.H3(title, className="text-xl font-semibold mb-2 text-accent"),
            html.Span([
                html.Span(className=f"animate-ping absolute inline-flex h-full w-full rounded-full opacity-75 {
                          'bg-green-400' if connected else 'bg-red-400'} "),
                html.Span(className=f"relative inline-flex rounded-full h-3 w-3 {
                          'bg-green-500' if connected else 'bg-red-500'}")
            ],
                className="relative flex h-3 w-3"
            )
        ], className="flex items-center justify-between"),
        html.P(accountType.replace('_', ' ').lower().capitalize(), className="text-2xl font-bold text-white"),
        html.P(mode, className="text-2xl font-bold text-white")
    ], className="bg-primary-foreground p-6 rounded-lg shadow-lg border border-accent")



from ..base_executor import BaseExecutor
from ...insight import InsightState


class CloseExhaustedInsightExecutor(BaseExecutor):
    """
    ### Executor for Closing Exhausted Insights
    This executor is used to close exhausted insights.

   :param strategy (BaseStrategy): The strategy instance

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.FILLED, "1.0")

    def run(self, insight):
        # Close the exhausted insight
        try:
            # check if the insight has not already been closed 
            if insight._closing:
                return self.returnResults(False, True, "Insight is being closed.")
            
            if self.STRATEGY.insights[insight.INSIGHT_ID].hasExhaustedTTL(True):
                return self.returnResults(False, True, "Insight closed due to being exhausted TTL.")
            return self.returnResults(True, True, "Insight has not expired.")
        except Exception as e:
            return self.returnResults(False, False, f"Error closing exhausted insight: {e}")



from multiprocessing.managers import BaseManager, ValueProxy, DictProxy


class SharedStrategyManager(BaseManager):
    pass


# Register the shared memory strategy class
SharedStrategyManager.register('get_strategy')
# Register the shared memory account data
SharedStrategyManager.register('get_account')
# Register the shared memory for the strategy mode
SharedStrategyManager.register('get_mode')
# Register the shared memory for the strategy assets
SharedStrategyManager.register('get_assets')
# Register the shared memory for the strategy positions
SharedStrategyManager.register('get_positions')
# Register the shared memory for the strategy insights
SharedStrategyManager.register('get_insights')
# Register the shared memory for the strategy history
SharedStrategyManager.register('get_history')
# Register the shared memory for the strategy metrics
SharedStrategyManager.register('get_metrics')
# Register the shared memory for the strategy time
SharedStrategyManager.register('get_time')




import abc
from typing import TYPE_CHECKING, List, Optional, override

from pandas import DataFrame
import pandas_ta as ta
import numpy as np

from OlympusTrader.broker.interfaces import IAsset, IQuote
from OlympusTrader.insight.insight import Insight


def get_BaseStrategy():
    from ..strategy.base_strategy import BaseStrategy
    return BaseStrategy


if TYPE_CHECKING:
    from ..strategy.base_strategy import BaseStrategy

class AlphaResults():
    """
    ### Alpha Results
    This class is used to store the results of the Alpha.
    """
    insight: Insight
    """Insight generated by the alpha."""
    success: bool
    """Boolean value indicating if the executor was ran successful or not."""
    alpha: str
    """Name of the executor."""
    message: str
    """Message indicating the result of the alpha."""

    def __init__(self, insight: Optional[Insight] = None, success: bool = True, message: str = None, alpha: str = None):
        self.insight = insight
        if insight:
            # If the executor passed, then it was successfully ran
            success = True
        self.success = success
        if self.success == False:
            # If the executor was not successfully ran, then it did not pass
            self.insight = None
        self.message = message
        self.alpha = alpha
class BaseAlpha(abc.ABC):
    """
    ### Abstract class for alpha implementations.
    Used to generate insights based on the strategy.
    """
    NAME: str
    """Name of the alpha model."""
    VERSION: str
    """Version of the alpha model."""

    STRATEGY: get_BaseStrategy
    """Reference to the strategy instance."""

    TA: List[dict] = []
    """List of technical analysis needed for the alpha model."""

    baseConfidenceModifierField: Optional[str] = None
    """Field to modify base confidence."""
    @abc.abstractmethod
    def __init__(self, strategy: get_BaseStrategy, name: str, version: str = "1.0", baseConfidenceModifierField: Optional[str] = None) -> None:
        self.NAME = name
        self.VERSION = version

        # Reference to the strategy instance
        self.STRATEGY = strategy
        if baseConfidenceModifierField:
            self.baseConfidenceModifierField = baseConfidenceModifierField


    @abc.abstractmethod
    def start(self):
        """Initialize the alpha model once at the start. data, etc. in the state of the Strategy."""
        pass

    @abc.abstractmethod
    def init(self,  asset: IAsset):
        """Initialize the alpha model for each assets. variables, data, etc. in the state of the Strategy."""
        pass

    @abc.abstractmethod
    def generateInsights(self, symbol: str) -> AlphaResults:
        """Generate insights based on the alpha model."""
        pass

    def registerAlpha(self):
        self.STRATEGY.ALPHA_MODELS.append(self)
        # Set the technical analysis needed for the alpha
        self._loadTa()

    def returnResults(self, insight: Optional[Insight] = None, success: bool = True, message: str = None) -> AlphaResults:
        return AlphaResults(insight, success, message, self.NAME)

    def get_history(self, symbol: str) -> DataFrame:
        return self.STRATEGY.history[symbol].loc[symbol]

    def get_latest_bar(self, symbol: str) -> DataFrame:
        return self.get_history(symbol).iloc[-1]

    def get_previos_bar(self, symbol: str) -> DataFrame:
        return self.get_history(symbol).iloc[-2]
    
    def get_latest_quote(self, insight: Insight) -> IQuote:
        return self.STRATEGY.broker.get_latest_quote(insight.ASSET)

    def get_baseConfidenceModifier(self, symbol: str):
        if self.baseConfidenceModifierField:
            baseConfidenceModifier = self.STRATEGY.state[self.baseConfidenceModifierField][symbol]
            if baseConfidenceModifier:
                return baseConfidenceModifier

        return 1
    def _loadTa(self):
        if self.TA:
            self.STRATEGY.add_ta(self.TA)



import datetime
from json import JSONEncoder
from typing import Any
import dash
from dash import dcc, html
from dash.dependencies import Input, Output
from uuid import UUID
import pandas as pd

# Import the SharedStrategyManager class
from OlympusTrader.broker.interfaces import IAccount, IAsset, IOrderSide, IPosition, ITradeUpdateEvent
from OlympusTrader.insight.insight import Insight, InsightState, StrategyDependantConfirmation
# from OlympusTrader.strategy.base_strategy import BaseStrategy
from OlympusTrader.strategy.interfaces import IStrategyMatrics, IStrategyMode
from OlympusTrader.ui.helper import sharedStrategyManager
from OlympusTrader.ui.helper.tradingViewHelper import history_to_trading_view_format
from OlympusTrader.ui.interfaces.store import STRATEGY_STORE_MAPPINGS, STRATEGY_SYNC_MAPPING

# FIXME: We will temporarily Override the JSoneEncoder to handle UUID objects. When we move to using data classes, we can remove this nasty hack.
old_default = JSONEncoder.default

def new_default(self, obj):
    if isinstance(obj, UUID):
        return str(obj)
    elif isinstance(obj, IOrderSide):
        return str(obj.value)
    elif isinstance(obj, InsightState):
        return str(obj)
    elif isinstance(obj, StrategyDependantConfirmation):
        return str(obj)
    elif isinstance(obj, IStrategyMode):
        return str(obj)
    elif isinstance(obj, StrategyDependantConfirmation):
        return str(obj)
    elif isinstance(obj, ITradeUpdateEvent):
        return str(obj)
    return old_default(self, obj)

JSONEncoder.default = new_default

# Initialize the SharedStrategyManager
global SSM_manager 
SSM_manager = sharedStrategyManager.get_shared_strategy_manager()


# External resources
external_stylesheets = [
    # "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.6.0/css/solid.min.css",
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.6.0/css/all.min.css"
    ]
external_scripts = ["https://cdn.tailwindcss.com"]

# Initialize the Dash app
app = dash.Dash(__name__,
                external_scripts=external_scripts,
                external_stylesheets=external_stylesheets,
                use_pages=True)

header = html.Header(
    className="text-white p-4 border-b border-accent",
    children=[
        html.Div(
            className="container mx-auto flex items-center justify-between",
            children=[
                html.Div(
                    className="flex items-center",
                    children=[
                        # TODO: Add logo
                        html.H1("OlympusTrader", className="text-2xl"),
                    ]
                ),
                html.Nav(
                    className="flex space-x-4",
                    children=[

                        dcc.Link(f"Overview", href="/", className="text-accent hover:text-white"),
                        html.A(
                            "Docs",
                            href="https://olympustrader.readthedocs.io/en/latest/?badge=latest",
                            target="_blank",
                            className="text-accent hover:text-white ml-4",
                            ),
                            html.A(
                            href="https://discord.gg/v2R5dcvssG",
                            target="_blank",
                            className="text-accent hover:text-white",
                            children=[
                                html.I(className="fab fa-discord fa-2x")  # Font Awesome Discord icon
                            ]
                        )

                        
                        # dcc.Link(f"{page['name']}", href=page["relative_path"]) for page in dash.page_registry.values()
                    ]
                )
            ]
        )
    ]
)

body = html.Main(
    children=dash.page_container
)

footer = html.Footer(
    className="bg-primary text-accent py-4 border-t border-accent",
    children=[
        html.Div(
            className="container mx-auto text-center",
            children=[
                html.P("© 2024 OlympusTrader. All rights reserved."),
                html.A(
                    href="https://github.com/isaac-diaby/OlympusTrader",
                    target="_blank",
                    className="text-accent hover:text-white ml-4 mt-1",
                    children=[
                        html.I(className="fab fa-github fa-2x")  # Font Awesome GitHub icon
                    ]
                ),
                # html.A(
                #     href="https://discord.gg/v2R5dcvssG",
                #     target="_blank",
                #     className="text-accent hover:text-white ml-4",
                #     children=[
                #         html.I(className="fab fa-discord fa-2x")  # Font Awesome Discord icon
                #     ]
                # )
            ]
        )
    ]
)

# Add the dcc.Interval component
interval_component = dcc.Interval(
    id=STRATEGY_SYNC_MAPPING.id,
    # interval=2*1000,  # Update every 2 seconds  - some works sometimes dont complete in 2 seconds so the data is not updated on the UI and the next updates come in
    interval=5*1000,  # Update every 5 seconds
    # interval=60*1000,  # Update every minute - for testing
    n_intervals=0
)

# Add a dcc.Store component to store the strategy data
# TODO: Might need to change the storage type to a  storage_type='local' or 'session'. This is becuase we can let the front end store the account balance history and calculate the profit/loss over time on the front end. in a local storage. This will reduce the amount of data that needs to be sent to the front end on page load.
strategy_data_store = [
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.account.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.mode.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.assets.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.positions.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.insights.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.history.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.metrics.id),
    dcc.Store(id=STRATEGY_STORE_MAPPINGS.time.id),
    ]

# Define the layout
app.layout = html.Div([header, body, *strategy_data_store, interval_component,
                      footer], className="min-h-screen bg-primary flex flex-col")

# Define the callback to update strategy data


@app.callback(
    [
        Output(STRATEGY_STORE_MAPPINGS.account.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.mode.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.assets.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.positions.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.insights.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.history.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.metrics.id, 'data'),
        Output(STRATEGY_STORE_MAPPINGS.time.id, 'data'),
    ],
    [
        Input(STRATEGY_SYNC_MAPPING.id, 'n_intervals'),
    ]
)
def update_strategy_data(n_intervals):
    # Fetch strategy data using the SharedStrategyManager

    global SSM_manager 
    if SSM_manager == None:
        SSM_manager = sharedStrategyManager.get_shared_strategy_manager()

    if n_intervals == None or SSM_manager == None:
        return ({},), (0,), ("Loading...",), ({},), ({},), (pd.DataFrame()) ({},), (None,)
    
    STRATEGY = SSM_manager.get_strategy()
    # Convert the objects to serializable formats
    ACCOUNT: IAccount = {k: v for k, v in SSM_manager.get_account().items()}
    ASSETS: dict[str, IAsset] = {k: v for k, v in SSM_manager.get_assets().items()}
    POSITIONS: dict[str, IPosition] = {k: v for k, v in SSM_manager.get_positions().items()}
    MODE: IStrategyMode = str(SSM_manager.get_mode().strip("'"))
    INSIGHTS: dict[str, Insight] = {k: v for k, v in SSM_manager.get_insights().items()}
    HISTORY = { k : history_to_trading_view_format(v) for k, v in SSM_manager.get_history().items() }
    METRICS: IStrategyMatrics =  {k: v for k, v in SSM_manager.get_metrics().items()}
    
    BROKER_TIME: datetime.datetime = int(str(SSM_manager.get_time()).split('.')[0])
    # print("Broker Time:", BROKER_TIME)



    return ACCOUNT, MODE, ASSETS, POSITIONS, INSIGHTS, HISTORY, METRICS, BROKER_TIME


# Run the server
if __name__ == '__main__':
    devMode = True
    if devMode:
        app.enable_dev_tools(
            debug=True,
            dev_tools_ui=True,
            dev_tools_serve_dev_bundles=True,
        )

    app.run(debug=devMode, threaded=devMode)




from typing import Optional
import numpy as np
import pandas as pd

from ..broker.interfaces import IOrderSide
from ..alpha.base_alpha import BaseAlpha
from ..insight.insight import Insight, StrategyDependantConfirmation


class RSIDiverganceAlpha(BaseAlpha):
    """
    ### RSI Divergance Alpha
    This alpha model generates insights based on RSI Divergance.

    :param strategy (BaseStrategy): The strategy instance
    :param local_window (int): The window to consider for local swing points
    :param divergance_window (int): The window to consider for divergance
    :param atrPeriod (int): The period for the ATR indicator
    :param rsiPeriod (int): The period for the RSI indicator
    :param baseConfidenceModifierField (str): The field to use for modifying the base confidence

    Author:
        @isaac-diaby

    """
    local_window: int
    divergance_window: int

    atrColumn: str
    rsiColumn: str

    def __init__(self, strategy, local_window=36, divergance_window=50, atrPeriod=14, rsiPeriod=14, baseConfidenceModifierField=None):
        super().__init__(strategy, "RSI_DIVERGANCE", "0.2", baseConfidenceModifierField)
        self.TA = [
            {"kind": 'atr', "length": atrPeriod},
            {"kind": 'rsi', "length": rsiPeriod, "scalar": 10}
        ]
        self.STRATEGY.warm_up = max(atrPeriod, rsiPeriod)
        self.atrColumn = f'ATRr_{atrPeriod}'
        self.rsiColumn = f'RSI_{rsiPeriod}'

        self.local_window = local_window
        self.divergance_window = divergance_window

    def start(self):
        self.STRATEGY.state['local_window'] = self.local_window
        self.STRATEGY.state['divergance_window'] = self.divergance_window

    def init(self, asset):
        pass

    def generateInsights(self, symbol):
        try:
            # v0.1
            # Compute Local Points of Control
            # self.computeLocalPointsOfControl(symbol)

            # Compute Local Swing Points
            self.computeLocalSwingPoints(symbol)

            # Check if we should even consider we have a divergance if we just closed on a pivot point (new high or low)
            if (np.isnan(self.STRATEGY.HISTORY[symbol]['higher_high'].iloc[-1]) and np.isnan(self.STRATEGY.HISTORY[symbol]['lower_low'].iloc[-1])):
                return self.returnResults()
            
            # Debugging
            # print(self.STRATEGY.HISTORY[symbol]
            #       [['higher_high', 'lower_low']].tail(2))

            # Compute RSI Divergance
            self.computeRSIDivergance(symbol)
            latestBar = self.get_latest_bar(symbol)
            previousBar = self.get_previos_bar(symbol)
            latestIATR = latestBar[self.atrColumn]
            baseConfidence = self.STRATEGY.baseConfidence

            # Modify Confidence based on baseConfidenceModifierField
            if (self.baseConfidenceModifierField):
                baseConfidence *= abs(self.get_baseConfidenceModifier(symbol))
            if baseConfidence <= 0:
                return self.returnResults(message="Base Confidence is 0.")

            # RSA Divergance Long
            # and marketState < 0):
            if (not np.isnan(latestBar['RSI_Divergance_Long'])):
                # print(f"Insight - {symbol}: Long Divergance: {latestBar['RSI_Divergance_Long']}")
                TP = self.STRATEGY.tools.dynamic_round(
                    (latestBar.close + (latestIATR*3.5)), symbol)
                SL = self.STRATEGY.tools.dynamic_round(
                    (latestBar.close - (latestIATR*1.5)), symbol)
                ENTRY = previousBar.high if (abs(
                    previousBar.high - latestBar.close) < latestIATR) else self.STRATEGY.tools.dynamic_round((latestBar.open+(.2*latestIATR)), symbol)
                # time to live unfilled
                TTLUF = self.STRATEGY.tools.calculateTimeToLive(
                    latestBar['close'], ENTRY, latestIATR)
                # time to live till take profit
                TTLF = self.STRATEGY.tools.calculateTimeToLive(
                    TP, ENTRY, latestIATR)

                return self.returnResults(Insight(IOrderSide.BUY, symbol,
                                                  self.NAME, self.STRATEGY.resolution, None, ENTRY, [TP], SL, baseConfidence, [StrategyDependantConfirmation.LRVCM], TTLUF, TTLF))
            # RSA Divergance Short
            # and marketState > 0):
            if (self.STRATEGY.assets[symbol]['shortable'] and not np.isnan(latestBar['RSI_Divergance_Short'])):
                # print(f"Insight - {symbol}: Short Divergance: {latestBar['RSI_Divergance_Short']}")
                TP = self.STRATEGY.tools.dynamic_round(
                    (latestBar.close - (latestIATR*3.5)), symbol)
                SL = self.STRATEGY.tools.dynamic_round(
                    (latestBar.close + (latestIATR*1.5)), symbol)
                ENTRY = previousBar.low if (abs(
                    previousBar.low - latestBar.close) < latestIATR) else self.STRATEGY.tools.dynamic_round((latestBar.open+(.2*latestIATR)), symbol)
                # time to live unfilled
                TTLUF = self.STRATEGY.tools.calculateTimeToLive(
                    latestBar['close'], ENTRY, latestIATR)
                # time to live till take profit
                TTLF = self.STRATEGY.tools.calculateTimeToLive(
                    TP, ENTRY, latestIATR)

                return self.returnResults(Insight(IOrderSide.SELL, symbol,
                                                  self.NAME, self.STRATEGY.resolution, None, ENTRY, [TP], SL, baseConfidence, [StrategyDependantConfirmation.LRVCM], TTLUF, TTLF))

            return self.returnResults()
        except Exception as e:
            return self.returnResults(success=False, message=str(e))
        
    # V0.2
    def computeLocalSwingPoints(self, symbol: str):
        window = self.STRATEGY.state['local_window']
        history = self.get_history(symbol)
        high_col = 'high'
        low_col = 'low'
        
        # Focus on the latest window of data
        recent_history = history.iloc[-window:]
        
        # Initialize columns in STRATEGY.HISTORY if they don't exist
        for col in ['higher_high', 'lower_low', 'lower_high', 'higher_low']:
            if col not in self.STRATEGY.HISTORY[symbol].columns:
                self.STRATEGY.HISTORY[symbol][col] = np.nan
        
        # Identify swing points
        higher_high = recent_history[high_col].iloc[-1] > recent_history[high_col].iloc[:-1].max()
        lower_low = recent_history[low_col].iloc[-1] < recent_history[low_col].iloc[:-1].min()
        lower_high = (recent_history[high_col].iloc[-1] < recent_history[high_col].iloc[:-1].max()) and \
                    (recent_history[high_col].iloc[-1] > recent_history[high_col].iloc[-2])
        higher_low = (recent_history[low_col].iloc[-1] > recent_history[low_col].iloc[:-1].min()) and \
                    (recent_history[low_col].iloc[-1] < recent_history[low_col].iloc[-2])
        
        # Update STRATEGY.HISTORY for the latest point
        latest_index = (symbol, recent_history.index[-1])
        
        if higher_high:
            self.STRATEGY.HISTORY[symbol].loc[latest_index, 'higher_high'] = recent_history[high_col].iloc[-1]
        if lower_low:
            self.STRATEGY.HISTORY[symbol].loc[latest_index, 'lower_low'] = recent_history[low_col].iloc[-1]
        if lower_high:
            self.STRATEGY.HISTORY[symbol].loc[latest_index, 'lower_high'] = recent_history[high_col].iloc[-1]
        if higher_low:
            self.STRATEGY.HISTORY[symbol].loc[latest_index, 'higher_low'] = recent_history[low_col].iloc[-1]
        
        return history
    

    def computeRSIDivergance(self, symbol: str):
        window = self.STRATEGY.state['divergance_window']
        # remove first 14 rows for RSI Warmup
        history = self.STRATEGY.HISTORY[symbol]
        # history = self.get_history(symbol)
        IRSI = history[self.rsiColumn]

        if 'RSI_Divergance_Long' not in history.columns:
            # Bullish Divergance - RSI is Increasing while price is Decreasing
            history['RSI_Divergance_Long'] = np.nan
        if 'RSI_Divergance_Short' not in history.columns:
            # Bearish divergence - RSI is Decreasing while price is Increasing
            history['RSI_Divergance_Short'] = np.nan


        if pd.notna(history['lower_low'].iloc[-1]):
            # only use local lows of point of control for reversal
            longPivot = history['lower_low'].dropna()
            # longPivot = history['local_min_poc'].dropna()
            lowerLowsPivots = longPivot.loc[longPivot.shift(1) > longPivot]

            for index, price in lowerLowsPivots[-1:-window:-1].items():
                _, *previousLocalPoC = longPivot.loc[index: (index[0], self.STRATEGY.resolution.add_time_increment(index[1], -2)): -1].items()
                if (len(previousLocalPoC) == 0):
                    continue
                lastLowIndex, lastPrice = previousLocalPoC[0]
                if (IRSI[lastLowIndex] < IRSI[index]):
                    self.STRATEGY.history[symbol].loc[index, [
                        'RSI_Divergance_Long']] = lastPrice-price

        if pd.notna(history['higher_high'].iloc[-1]):
            # only use local maximas of point of control for reversal
            shortPivot = history['higher_high'].dropna()
            # shortPivot = history['local_max_poc'].dropna()
            higherHighsPivots = shortPivot.loc[shortPivot.shift(1) < shortPivot]

            for index, price in higherHighsPivots[-1:-window:-1].items():
                _, *previousLocalPoC = shortPivot.loc[index: (index[0], self.STRATEGY.resolution.add_time_increment(index[1], -2)): -1].items()
                if (len(previousLocalPoC) == 0):
                    continue
                lastHighIndex, lastPrice = previousLocalPoC[0]
                if (IRSI[lastHighIndex] > IRSI[index]):
                    self.STRATEGY.history[symbol].loc[index, [
                        'RSI_Divergance_Short']] = price-lastPrice

        return history


    # V0.1
    # def computeLocalSwingPoints(self, symbol: str):
    #     window = self.STRATEGY.state['local_window']
    #     history = self.get_history(symbol)
    #     high_col = 'high'
    #     low_col = 'low'
        
    #     # Create a new DataFrame to store the results
    #     swing_points = pd.DataFrame(index=history.index, columns=['higher_high', 'lower_low', 'lower_high', 'higher_low'])
        
    #     for i in range(window, len(history) - window):
    #         current_index = history.index[i]
            
    #         # Check for higher high
    #         if (history[high_col].iloc[i] > history[high_col].iloc[i-window:i].max() and 
    #             history[high_col].iloc[i] > history[high_col].iloc[i+1:i+window+1].max()):
    #             swing_points.loc[current_index, 'higher_high'] = history[high_col].iloc[i]
            
    #         # Check for lower low
    #         if (history[low_col].iloc[i] < history[low_col].iloc[i-window:i].min() and 
    #             history[low_col].iloc[i] < history[low_col].iloc[i+1:i+window+1].min()):
    #             swing_points.loc[current_index, 'lower_low'] = history[low_col].iloc[i]
            
    #         # Check for lower high
    #         if (history[high_col].iloc[i] < history[high_col].iloc[i-window:i].max() and 
    #             history[high_col].iloc[i] > history[high_col].iloc[i+1:i+window+1].max()):
    #             swing_points.loc[current_index, 'lower_high'] = history[high_col].iloc[i]
            
    #         # Check for higher low
    #         if (history[low_col].iloc[i] > history[low_col].iloc[i-window:i].min() and 
    #             history[low_col].iloc[i] < history[low_col].iloc[i+1:i+window+1].min()):
    #             swing_points.loc[current_index, 'higher_low'] = history[low_col].iloc[i]
        
    #     # Handle the last window of data
    #     last_index = history.index[-1]
    #     last_window = history.iloc[-window:]
        
    #     if last_window[high_col].iloc[-1] == last_window[high_col].max():
    #         swing_points.loc[last_index, 'higher_high'] = last_window[high_col].iloc[-1]
        
    #     if last_window[low_col].iloc[-1] == last_window[low_col].min():
    #         swing_points.loc[last_index, 'lower_low'] = last_window[low_col].iloc[-1]
        
    #     # Update the STRATEGY.HISTORY DataFrame
    #     for col in swing_points.columns:
    #         self.STRATEGY.HISTORY[symbol][col] = swing_points[col]
        
    #     return history

    # def computeLocalPointsOfControl(self, symbol: str):
    #     window = self.STRATEGY.state['local_window']
    #     history = self.get_history(symbol)
    #     viewColumn = 'close'
    #     self.STRATEGY.HISTORY[symbol].loc[[symbol], ['local_max_poc']] = history[viewColumn][(
    #         history[viewColumn].shift(window) < history[viewColumn]) & (history[viewColumn].shift(-window) < history[viewColumn]
    #                                                                     )]
    #     self.STRATEGY.HISTORY[symbol].loc[[symbol], ['local_min_poc']] = history[viewColumn][(
    #         history[viewColumn].shift(window) > history[viewColumn]) & (history[viewColumn].shift(-window) > history[viewColumn]
    #                                                                     )]
    #     return history

    # def computeRSIDivergance(self, symbol: str):
    #     window = self.STRATEGY.state['divergance_window']
    #     # remove first 14 rows for RSI Warmup
    #     history = self.get_history(symbol)
    #     IRSI = history[self.rsiColumn]

    #     # Bullish Divergance - RSI is Increasing while price is Decreasing
    #     self.STRATEGY.history[symbol].loc[[symbol],
    #                                       ['RSI_Divergance_Long']] = np.nan

    #     # only use local lows of point of control for reversal
    #     longPivot = history['lower_low'].dropna()
    #     # longPivot = history['local_min_poc'].dropna()
    #     lowerLowsPivots = longPivot.loc[longPivot.shift(1) > longPivot]

    #     for index, price in lowerLowsPivots[-1:-window:-1].items():
    #         _, *previousLocalPoC = longPivot.loc[index: (
    #             index[0], self.STRATEGY.resolution.add_time_increment(index[1], -2)): -1].items()
    #         if (len(previousLocalPoC) == 0):
    #             continue
    #         lastLowIndex, lastPrice = previousLocalPoC[0]
    #         if (IRSI.loc[lastLowIndex] < IRSI.loc[index]):
    #             self.STRATEGY.history[symbol].loc[index, [
    #                 'RSI_Divergance_Long']] = lastPrice-price

    #     # Bearish divergence - RSI is Decreasing while price is Increasing
    #     self.STRATEGY.history[symbol].loc[[symbol],
    #                                       ['RSI_Divergance_Short']] = np.nan

    #     # only use local maximas of point of control for reversal
    #     shortPivot = history['higher_high'].dropna()
    #     # shortPivot = history['local_max_poc'].dropna()
    #     higherHighsPivots = shortPivot.loc[shortPivot.shift(1) < shortPivot]

    #     for index, price in higherHighsPivots[-1:-window:-1].items():
    #         _, *previousLocalPoC = shortPivot.loc[index: (
    #             index[0], self.STRATEGY.resolution.add_time_increment(index[1], -2)): -1].items()
    #         if (len(previousLocalPoC) == 0):
    #             continue
    #         lastHighIndex, lastPrice = previousLocalPoC[0]
    #         if (IRSI.loc[lastHighIndex] > IRSI.loc[index]):
    #             self.STRATEGY.history[symbol].loc[index, [
    #                 'RSI_Divergance_Short']] = price-lastPrice

    #     return history



from ..broker.interfaces import IOrderSide
from .base_alpha import BaseAlpha
from ..insight.insight import Insight, StrategyDependantConfirmation, StrategyTypes


class TestEntryAlpha(BaseAlpha):
    """
    ### Test Entry Alpha
    This alpha model generates insights if there is no other insight.
    This is intended for testing purposes.

    :param strategy (BaseStrategy): The strategy instance
    :param atrPeriod (int): The period for the ATR indicator
    :param limitEntries (bool): Whether to limit entries
    :param maxSpawn (int): The maximum number of insights that can spawn at the same time
    :param baseConfidenceModifierField (str): The field to use for modifying the base confidence

    Author:
        @isaac-diaby
    """
    atrColumn: str
    limitEntry: bool
    maxSpawn: int

    def __init__(self, strategy,  atrPeriod=14, limitEntries=False, maxSpawn=1, baseConfidenceModifierField=None):
        super().__init__(strategy, "TEST_ENTRY", "1.0", baseConfidenceModifierField)
        self.TA = [
            {"kind": 'atr', "length": atrPeriod},
        ]
        self.atrColumn = f'ATRr_{atrPeriod}'

        self.STRATEGY.warm_up = atrPeriod
        self.limitEntry = limitEntries
        self.maxSpawn = maxSpawn

    def start(self):
        pass

    def init(self, asset):
        pass

    def generateInsights(self, symbol):
        try:
            latestBar = self.get_latest_bar(symbol)
            previousBar = self.get_previos_bar(symbol)
            latestIATR = latestBar[self.atrColumn]
            baseConfidence = self.STRATEGY.baseConfidence
            # Modify Confidence based on baseConfidenceModifierField
            if (self.baseConfidenceModifierField):
                baseConfidence *= abs(self.get_baseConfidenceModifier(symbol))

            if baseConfidence <= 0:
                return self.returnResults(message="Base Confidence is 0.")

            if len(self.STRATEGY.insights) >= self.maxSpawn:
                return self.returnResults(message="Insight already exists.")

            # Generate Test Entry
            if (latestBar.close > latestBar.open):
                TP = self.STRATEGY.tools.dynamic_round(
                    latestBar.close + (latestIATR*2), symbol)
                SL = self.STRATEGY.tools.dynamic_round(
                    latestBar.close - (latestIATR*1.5), symbol)
                ENTRY = self.STRATEGY.tools.dynamic_round(
                    latestBar['close'], symbol) if self.limitEntry else None

                return self.returnResults(insight=Insight(
                    side=IOrderSide.BUY,
                    symbol=symbol,
                    strategyType=StrategyTypes.TEST,
                    tf=self.STRATEGY.resolution,
                    quantity=None,
                    limit_price=ENTRY,
                    TP=[TP],
                    SL=SL,
                    confidence=baseConfidence,
                    periodUnfilled=5,
                    periodTillTp=10
                ))
            else:
                if self.STRATEGY.assets[symbol]["shortable"]:
                    TP = self.STRATEGY.tools.dynamic_round(
                        latestBar.close - (latestIATR*2), symbol)
                    SL = self.STRATEGY.tools.dynamic_round(
                        latestBar.close + (latestIATR*1.5), symbol)
                    ENTRY = self.STRATEGY.tools.dynamic_round(
                        latestBar['close'], symbol) if self.limitEntry else None

                    return self.returnResults(insight=Insight(
                        side=IOrderSide.SELL,
                        symbol=symbol,
                        strategyType=StrategyTypes.TEST,
                        tf=self.STRATEGY.resolution,
                        quantity=None,
                        limit_price=ENTRY,
                        TP=[TP],
                        SL=SL,
                        confidence=baseConfidence,
                        periodUnfilled=5,
                        periodTillTp=10
                    ))
                else:
                    return self.returnResults(message="Asset is not shortable.")
            #     return self.returnResults(message="Insight already exists.")
        except Exception as e:
            return self.returnResults(success=False, message=str(e))



from dataclasses import dataclass
from datetime import datetime
from typing import Literal, Optional, TypedDict, Required, List
from enum import Enum
import pandas_ta as ta
from numpy import divide


from ..utils.timeframe import ITimeFrame


class IMarketDataStream(TypedDict):
    symbol: Required[str]
    exchange: str
    time_frame: Required[ITimeFrame]
    feature: Optional[str] = None
    asset_type: Literal['stock', 'crypto'] = 'crypto'
    type: Required[Literal['trade', 'quote', 'bar', 'news']] = 'bar'
    stored: Optional[bool] = False
    stored_path: Optional[str] = None
    start: Optional[datetime] = None
    end: Optional[datetime] = None
    applyTA: Optional[bool] = False
    TA: Optional[ta.Strategy] = None


class IBacktestingConfig(TypedDict):
    preemptiveTA: Optional[bool] = False


class IStrategyMode(Enum):
    BACKTEST = 'Backtest'
    LIVE = 'Live'


@dataclass
class IStrategyMatrics:
    start_date: Optional[datetime] = None
    """ The start date """
    end_date: Optional[datetime] = None
    starting_cash: float = 0.0
    """ The starting cash """

    total_open: int = 0
    """ The total number of open trades """
    total_closed: int = 0
    """ The total number of closed trades """
    total_wins: int = 0
    """ The total number of winning trades """
    total_losses: int = 0
    """ The total number of losing trades """
    total_profit: float = 0.0
    """ The total profit """
    total_loss: float = 0.0
    """ The total loss """
    total_pnl: float = 0.0
    """ The total profit and loss """

    # DERIVED METRICS
    win_rate: float = 0.0
    """ The total win rate """
    avg_win: float = 0.0
    """ The average win """
    avg_loss: float = 0.0
    """ The average loss """

    # TODO: Add the following metrics
    # total_risk: float = 0.0
    # """ The total risk - market value of the trades """
    # total_risk_reward_ratio: float = 0.0
    # """ The total risk reward ratio """
    # total_fees: float = 0.0
    # """ The total fees """
    # total_profit_factor: float = 0.0
    # """ The total profit factor """
    # total_sharpe_ratio: float = 0.0
    # """ The total sharpe ratio """

    def updateStart(self, start_date: datetime, starting_cash: float):
        """ Update the start date and starting cash """
        self.start_date = start_date
        self.starting_cash = starting_cash

    def updateDerivedMetrics(self):
        """ Update the derived metrics """
        if self.total_wins == 0 or self.total_closed == 0:
            self.win_rate = 0.0
        else:
            self.win_rate = round(divide(self.total_wins,  self.total_closed),2)

        if self.total_profit == 0 or self.total_wins == 0:
            self.avg_win = 0.0
        else:
            self.avg_win = divide(self.total_profit, self.total_wins)

        if self.total_loss == 0 or self.total_losses == 0:
            self.avg_loss = 0.0
        else:
            self.avg_loss = divide(self.total_loss, self.total_losses)

    def positionOpened(self):
        """ Update the metrics when a position is opened """
        self.total_open += 1

    def positionClosed(self, pnl: float):
        """ Update the metrics when a position is closed """
        if pnl > 0:
            self.total_wins += 1
            self.total_profit += pnl
        else:
            self.total_losses += 1
            self.total_loss += pnl
        self.total_pnl += pnl
        self.total_closed += 1
        self.updateDerivedMetrics()



from dash import html
def MatricItem(label: str, value: str):
    return html.Div(
        # className='matric-item',
        children=[
            html.H4(label, className="text-accent text-sm"),
            html.P(value, className="text-white text-lg font-semibold")
        ]
    )


from typing import Optional
from dash import html, dcc, callback, no_update
from dash.dependencies import Input, Output, State, MATCH
from dash.exceptions import PreventUpdate


from OlympusTrader.broker.interfaces import IOrderSide, IPosition
from OlympusTrader.insight.insight import IInsight, InsightState
from OlympusTrader.ui.interfaces.store import STRATEGY_STORE_MAPPINGS


def tradeRow(insight: IInsight, position: Optional[IPosition]):
    # TODO: change this to use the currency symbol from the account
    currencySymbol = "£"

    def calculateProfitLoss(insight):
        if position == None:
            return 0

        # Check if the insight has been filled
        if insight["state"] != InsightState.FILLED.value:
            return 0

        currentPrice = insight["close_price"] if insight["close_price"] != None else position["current_price"]
        match insight["side"]:
            case IOrderSide.BUY.value:
                return (currentPrice - insight["limit_price"]) * insight.get("quantity", 0)
            case IOrderSide.SELL.value:
                return (insight["limit_price"] - currentPrice) * insight.get("quantity", 0)
            case _:
                print("Invalid side value - tradeRow()")
                return 0

    profitLoss = calculateProfitLoss(insight)
    return html.Tr([
        html.Td(
            insight["insight_id"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["symbol"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["side"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["type"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["strategy"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["contracts"] if insight["useContractSize"] else insight["quantity"], className="px-6 py-4  whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["limit_price"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(
            insight["close_price"], className="px-6 py-4 whitespace-nowrap text-sm text-white"),
        html.Td(f"{currencySymbol}{profitLoss}",
                className=f"px-6 py-4  whitespace-nowrap text-sm {'text-green-500' if (profitLoss > 0) else 'text-red-500' if (profitLoss < 0) else 'text-white'}")
    ], className="hover:bg-primary-light cursor-pointer")


def tradeTable(index_id, FILTERS={}):

    return html.Div([
        html.Table([
            html.Thead([
                html.Tr([
                   html.Th(
                       "Insight ID", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Asset", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Side", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Type", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Strategy", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "size", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Entry", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Exit", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   html.Th(
                       "Profit/Loss", className="px-6 py-3 text-left text-xs font-medium text-accent uppercase tracking-wider"),
                   ])
            ], className="bg-primary-light"),
            html.Tbody(id={'type': 'trades-table-body', 'index': index_id},
                       className="bg-primary-foreground divide-y divide-accent"),
            dcc.Store(id={'type': 'trades-table-body-filter',
                      'index': index_id}, data=FILTERS)
        ], className="w-full")

    ], className="bg-primary-foreground overflow-x-auto shadow rounded-lg overflow-hidden border border-accent")


@callback(Output({'type': 'trades-table-body', 'index': MATCH}, 'children'),
          Input(STRATEGY_STORE_MAPPINGS.insights.id, 'data'),
          Input(STRATEGY_STORE_MAPPINGS.positions.id, 'data'),
          [
              State({'type': 'trades-table-body-filter', 'index': MATCH}, 'data'),
              State({'type': 'trades-table-body', 'index': MATCH}, 'children')
])
def update_trade_table_body(insights, positions, filter, children):
    def filterInsights(insights, filters: Optional[dict]):
        filterdInsights = []

        if filter == None:
            return insights

        for insight in insights.values():

            if all([(insight.get(key) == value) for key, value in filters.items()]):
                # Match all of the filters
                filterdInsights.append(insight)

        return filterdInsights

    if insights == None:
        raise PreventUpdate

    if filter == None:
        raise PreventUpdate

    entries = filterInsights(insights, filter)

    # If there are no entries and there are children, return no update
    if children != None and len(entries) == 0:
        return []

    # If there are no entries and no children, return no update
    if len(entries) == 0:
        return no_update

    # If there are entries, return the trade rows and update the table
    return [tradeRow(insight, (positions.get(insight["symbol"], None) if insight["symbol"] in positions else positions.get(insight["order_id"], None) )) for insight in entries]



from dataclasses import dataclass


@dataclass
class STORE_MAP():
    id: str
@dataclass
class STRATEGY_STORE_MAPPINGS():
    account = STORE_MAP("strategy-data-store-account")
    mode = STORE_MAP("strategy-data-store-mode")
    assets = STORE_MAP("strategy-data-store-assets")
    positions = STORE_MAP("strategy-data-store-positions")
    insights = STORE_MAP("strategy-data-store-insights")
    history = STORE_MAP("strategy-data-store-history")
    metrics = STORE_MAP("strategy-data-store-metrics")
    time = STORE_MAP("strategy-data-store-time")

@dataclass
class STRATEGY_SYNC_MAPPING(STORE_MAP):
    id = "strategy-sync-store"


import abc
import asyncio
from dataclasses import asdict
import functools
import os
from pathlib import Path
from threading import BrokenBarrierError
from time import sleep
from typing import Any, List, Optional, Literal
from uuid import uuid4, UUID
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
import datetime
import nest_asyncio
import timeit
from collections import deque
import logging

import pandas_ta as ta
import pytz
from tqdm import tqdm
from vectorbt.portfolio import Portfolio

from ..broker.base_broker import BaseBroker
from ..broker.interfaces import IOrderSide, ISupportedBrokers, ITradeUpdateEvent, IAsset, IAccount, IPosition, IOrder

from .sharedmemory import SharedStrategyManager
from .interfaces import IBacktestingConfig, IMarketDataStream, IStrategyMatrics, IStrategyMode
from ..insight.insight import Insight, InsightState
from ..utils.timeframe import ITimeFrame, ITimeFrameUnit
from ..utils.types import AttributeDict
from ..utils.tools import ITradingTools

from ..alpha.base_alpha import BaseAlpha
from ..insight.executors.base_executor import BaseExecutor
# from ..ui.base_ui import Dashboard
import warnings
warnings.simplefilter(action="ignore", category=FutureWarning)


class BaseStrategy(abc.ABC):
    STRATEGY_ID: UUID = uuid4()
    NAME: str = "BaseStrategy"
    BROKER: BaseBroker
    ACCOUNT: Optional[IAccount] = None
    POSITIONS: dict[str, IPosition] = {}
    ORDERS: Optional[dict[str, IOrder]] = {}
    HISTORY: dict[str, pd.DataFrame] = {}
    INSIGHTS: dict[UUID, Insight] = {}
    UNIVERSE: dict[str, IAsset] = {}
    RESOLUTION = ITimeFrame(5, ITimeFrameUnit.Minute)
    STREAMS: List[IMarketDataStream] = []
    VARIABLES: AttributeDict
    MODE: IStrategyMode
    WITHUI: bool = True
    SSM: Optional[SharedStrategyManager] = None
    # DASHBOARD: Dashboard = None
    tradeOnFeatureEvents: bool = False

    TOOLS: Optional[ITradingTools] = None

    # DEBUG
    VERBOSE: int = 0

    _Running = False

    # ALPHA MODELS
    ALPHA_MODELS: List[BaseAlpha] = []
    """Alpha models to be used in the strategy"""

    # Insight Executors Models
    INSIGHT_EXECUTORS: dict[InsightState, deque[BaseExecutor]] = {
        InsightState.NEW: deque([]),
        InsightState.EXECUTED: deque([]),
        InsightState.FILLED: deque([]),
        InsightState.CLOSED: deque([]),
        InsightState.REJECTED: deque([]),
        InsightState.CANCELED: deque([])
    }
    """Insight Executors Models"""

    # Strategy Execution Parameters
    TaStrategy: ta.Strategy = None
    WARM_UP: int = 0
    execution_risk: float = 0.01  # 1% of account per trade
    minRewardRiskRatio: float = 2.0  # 2:1 Reward to Risk Ratio minimum
    baseConfidence: float = 0.1  # Base Confidence level for the strategy
    shouldClosePartialFilledIfCancelled: bool = True
    """Insights that are partially filled and are cancelled should be closed if the insight is cancelled"""
    insightRateLimit: int = 1

    BACKTESTING_CONFIG: IBacktestingConfig = IBacktestingConfig(
        preemptiveTA=False)
    """Backtesting configuration"""
    BACKTESTING_RESULTS: dict[str, Portfolio] = {}
    """Backtesting results"""

    MATRICS: IStrategyMatrics = IStrategyMatrics()

    @abc.abstractmethod
    def __init__(self, broker: BaseBroker, variables: AttributeDict = AttributeDict({}), resolution: ITimeFrame = ITimeFrame(1, ITimeFrameUnit.Minute), verbose: int = 0, ui: bool = True, mode:
                 IStrategyMode = IStrategyMode.LIVE, tradeOnFeatureEvents: bool = False) -> None:
        """Abstract class for strategy implementations."""
        self.NAME = self.__class__.__name__
        self.MODE = mode
        self.WITHUI = ui
        self.VARIABLES = variables
        self.BROKER = broker
        self.TOOLS = ITradingTools(self)
        self.VERBOSE = verbose  # TODO: Log Levels should be an derived from env file
        assert ITimeFrame.validate_timeframe(
            resolution.amount, resolution.unit), 'Resolution must be a valid timeframe'
        self.RESOLUTION = resolution
        self.tradeOnFeatureEvents = tradeOnFeatureEvents
        # set the UI shared memory sever
        if self.WITHUI:
            self._startUISharedMemory()

        # Set up TA Strategy
        self.TaStrategy = ta.Strategy(
            name=self.NAME, description="Olympus Trader Framework", ta=[])

        self.start()
        # Load the universe
        self._loadUniverse()

        # Set backtesting configuration
        if self.MODE == IStrategyMode.BACKTEST:
            # check if the broker is paper
            assert self.BROKER.NAME == ISupportedBrokers.PAPER, 'Backtesting is only supported with the paper broker'
            # # change the broker to feed to backtest mode
            # self.BROKER.feed = IStrategyMode.BACKTEST
            pass

        try:
            self.ACCOUNT = self.BROKER.get_account()
            self.POSITIONS = self.BROKER.get_positions()
            self.ORDERS = self.BROKER.get_orders()

            # Initialise the strategy metrics
            self.MATRICS.updateStart((pd.Timestamp(self.BROKER.get_current_time, tz='UTC')-datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)).total_seconds(), self.ACCOUNT.equity)

        except Exception as e:
            print(f'Failed to get account info from the broker {e}')
            pass

    @abc.abstractmethod
    def start(self):
        """ Start the strategy. This method is called once at the start of the strategy."""
        pass

    @abc.abstractmethod
    def init(self, asset: IAsset):
        """ Initialize the strategy. This method is called once before the start of the strategy on every asset in the universe."""

    @abc.abstractmethod
    def universe(self) -> set[str]:
        """ Used to generate the stock in play. """
        pass

    @abc.abstractmethod
    def on_bar(self, symbol: str, bar: dict):
        """ Called once per bar. open, high, low, close, volume """
        print('IS THIS WORKING, Add on_bar Function? ->  async def on_bar(self, bar):')
        pass

    @abc.abstractmethod
    def generateInsights(self, symbol: str):
        """Called once per bar to generate insights. """
        print('IS THIS WORKING, Add generateInsights Function? -> def generateInsights(self, symbol: str):')
        pass

    @abc.abstractmethod
    def executeInsight(self, insight: Insight):
        """ Called for each active insight in the strategy.
        it allows you to conrol the execution of the insight and manage the order.
        """
        print('IS THIS WORKING, Add executeInsight Function? ->  async def executeInsight(self, symbol: str):')
        match insight.state:
            # case InsightState.NEW:
            #     pass
            # case InsightState.EXECUTED:
            #     pass
            # case InsightState.FILLED:
            #     pass
            # case InsightState.CLOSED:
            #     pass
            # case InsightState.REJECTED:
            #     pass
            # case InsightState.CANCELED:
            #     pass
            case _:
                print(
                    'Implement the insight state in the executeInsight function:', insight.state)
                pass

    @abc.abstractmethod
    def teardown(self):
        """ Called once at the end of the strategy. """
        print("Tear Down: Closing all positions")
        self.BROKER.close_all_positions()

    def _start(self):
        """Start the alpha models."""
        # assert callable(self.start), 'start must be a callable function'
        # print("Running start function")

        for alpha in self.ALPHA_MODELS:
            alpha.start()
        # self.start()

    def _init(self, asset: IAsset):
        """ Initialize the strategy. """
        assert callable(self.init), 'init must be a callable function'
        for alpha in self.ALPHA_MODELS:
            alpha.init(asset)
        self.init(asset)

    def _generateInsights(self, symbol: str):
        """ Generate insights for the strategy. """
        assert callable(
            self.generateInsights), 'generateInsights must be a callable function'
        for alpha in self.ALPHA_MODELS:
            result = alpha.generateInsights(symbol)
            if result.success:
                if result.insight is not None:
                    self.add_insight(result.insight)
            else:
                print(f"Alpha {result.alpha} Failed: {result.message}")

        self.generateInsights(symbol)

    def run(self):
        """Starts the strategy."""
        nest_asyncio.apply()
        self._Running = True

        self._start_strategy()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        try:
            with ThreadPoolExecutor(max_workers=5, thread_name_prefix="OlympusTrader") as pool:
                self._start_streams_and_listeners(loop, pool)
                self._run_strategy(loop, pool)
        except KeyboardInterrupt as e:
            print("Interrupted execution by user:", e)
            self._handle_close_streams(loop, pool)
        finally:
            self.run_teardown(loop)

    def _start_strategy(self):
        """Set up the strategy and check if the universe is empty."""
        self._start()
        assert len(self.UNIVERSE) > 0, 'Universe is empty'

    def _start_streams_and_listeners(self, loop, pool):
        """Start the trading data streams, market data streams, UI shared memory server, and insight listener."""
        self.TRADE_STREAM = loop.run_in_executor(
            pool, self.BROKER.startTradeStream, self._on_trade_update)
        self.MARKET_DATA_STREAM = loop.run_in_executor(
            pool, self.BROKER.streamMarketData, self._on_bar, self.STREAMS)
        self.INSIGHT_STREAM = loop.run_in_executor(pool, self._insightListener)

        if self.WITHUI:
            server = self.SSM.get_server()
            # Run the UI Shared Memory Server
            self.UI_SSM_STREAM = loop.run_in_executor(
                pool, server.serve_forever)
            from OlympusTrader.ui.__main__ import app
            # FIXME: Enabling dev tools is not working in a pool executor for some reason
            # if self.VERBOSE > 0:
            #     # Enable the dev tools
            #     app.enable_dev_tools(
            #         debug=True,
            #         dev_tools_ui=True,
            #         dev_tools_serve_dev_bundles=True,
            #         dev_tools_hot_reload=True
            #     )

            #     # Run the UI Dashboard Server in debug mode
            #     self.UI_STREAM = loop.run_in_executor(pool,  functools.partial(app.run, debug=True))
            # else:
            #     log = logging.getLogger('werkzeug')
            #     log.setLevel(logging.ERROR)
            #     # Run the UI Dashboard Server in production mode
            #     self.UI_STREAM = loop.run_in_executor(pool, functools.partial(app.run, host='0.0.0.0', threaded=True))
                # Run the UI Dashboard Server in production mode
            log = logging.getLogger('werkzeug')
            log.setLevel(logging.ERROR)
            
            self.UI_STREAM = loop.run_in_executor(pool, functools.partial(app.run, host='0.0.0.0', threaded=True))
            print('UI Shared Memory Server Started')

    def _run_strategy(self, loop, pool):
        """Run the strategy in backtest mode or live mode."""
        if self.MODE == IStrategyMode.BACKTEST:
            self._run_backtest()
        else:
            loop.run_forever()

        self.teardown()

    def _run_backtest(self):
        """Run the strategy in backtest mode."""
        while not self.BROKER.RUNNING_MARKET_STREAM or not self.BROKER.RUNNING_TRADE_STREAM:
            if self.BROKER.RUNNING_MARKET_STREAM and self.BROKER.RUNNING_TRADE_STREAM:
                break
        if self.VERBOSE > 0:
            print(f"Running Backtest - {datetime.datetime.now()}")
            start_time = timeit.default_timer()

        while self.BROKER.RUNNING_MARKET_STREAM and self.BROKER.RUNNING_TRADE_STREAM:
            if not self.BROKER.RUNNING_MARKET_STREAM or not self.BROKER.RUNNING_TRADE_STREAM:
                break

        if self.VERBOSE > 0:
            print('Backtest Completed:', timeit.default_timer() - start_time)
        # self.BACKTESTING_RESULTS = self.BROKER.get_VBT_results(self.resolution)
        # self.saveBacktestResults()

        raise KeyboardInterrupt("Backtest Completed")

    def _handle_close_streams(self, loop, pool):
        """Handle interruptions and ensure all streams and listeners are closed."""
        """Handle interruptions and ensure all streams and listeners are closed."""
        try:
            self.teardown()

            # Cancel all running tasks
            for task in asyncio.all_tasks(loop):
                task.cancel()

            # Cancel individual streams
            if hasattr(self, 'TRADE_STREAM'):
                self.TRADE_STREAM.cancel("From Main: Trade Stream Cancel")
            if hasattr(self, 'MARKET_DATA_STREAM'):
                self.MARKET_DATA_STREAM.cancel(
                    "From Main: Market Stream Cancel")
            if hasattr(self, 'INSIGHT_STREAM'):
                self.INSIGHT_STREAM.cancel()

            # Cancel UI streams if UI is enabled
            if self.WITHUI:
                if hasattr(self, 'UI_STREAM'):
                    self.UI_STREAM.cancel("From Main: UI Stream Cancel")
                if hasattr(self, 'UI_SSM_STREAM'):
                    self.UI_SSM_STREAM.cancel(
                        "From Main: UI SSM Stream Cancel")

            # Shutdown the thread pool
            pool.shutdown(wait=False, cancel_futures=True)
            # Stop the event loop
            loop.stop()

        except Exception as e:
            print(f"Error during interruption handling: {e}")
        finally:
            # Ensure the broker streams are closed
            asyncio.run(self.BROKER.closeTradeStream())
            asyncio.run(self.BROKER.closeStream(self.STREAMS))
            self._Running = False

    def run_teardown(self, loop):
        """Clean up resources and save backtest results if applicable."""
        asyncio.run(self.BROKER.closeTradeStream())
        asyncio.run(self.BROKER.closeStream(self.STREAMS))
        loop.stop()
        self._Running = False

        if self.MODE == IStrategyMode.BACKTEST and len(self.BACKTESTING_RESULTS) == 0:
            self.BACKTESTING_RESULTS = self.BROKER.get_VBT_results(
                self.resolution)
            self.saveBacktestResults()
            print("Simulation Account", self.BROKER.Account)

        exit(0)

    def saveBacktestResults(self):
        """ Save the backtest results."""
        if self.MODE != IStrategyMode.BACKTEST:
            print("Backtest results can only be saved in backtest mode")
            return
        # check if there are backtest results
        if len(self.BACKTESTING_RESULTS) == 0:
            print("No backtest results to save")
            return

        # Save the backtest results
        for symbol in tqdm(self.BACKTESTING_RESULTS.keys(), desc="Saving Backtest Results"):
            save_path = Path(f"backtests/{self.NAME}/{self.STRATEGY_ID}")
            save_path.mkdir(parents=True, exist_ok=True)
            path = f"{save_path}/{symbol}-{self.resolution}-backtest"
            if (self.BACKTESTING_RESULTS.get(symbol)):
                self.BACKTESTING_RESULTS[symbol].save(path)
                self.BACKTESTING_RESULTS[symbol].plot().show()
                print("Backtesting results saved for", symbol, "at", path)
            else:
                print("No backtesting results found for", symbol)

    def _startUISharedMemory(self):
        """ Starts the UI shared memory."""
        if not self.WITHUI:
            print('UI is not enabled')
            return
        try:
            assert os.getenv(
                "SSM_PASSWORD"), 'SSM_PASSWORD not found in environment variables'
            SharedStrategyManager.register(
                'get_strategy', callable=lambda: self)
            SharedStrategyManager.register(
                'get_account', callable=lambda: {str(key): v for key, v in asdict(self.account).items()})
            SharedStrategyManager.register(
                'get_mode', callable=lambda: self.MODE.value)
            SharedStrategyManager.register(
                'get_assets', callable=lambda: self.assets)
            SharedStrategyManager.register(
                'get_positions', callable=lambda: self.positions)
            SharedStrategyManager.register(
                'get_insights', callable=lambda: {str(key): asdict(insight.dataclass) for key, insight in self.INSIGHTS.items()})
            # SharedStrategyManager.register(
            #     'get_history', callable=lambda: { k : history_to_trading_view_format(v) for k, v in self.history.items() })
            SharedStrategyManager.register(
                'get_history', callable=lambda: self.history)
            SharedStrategyManager.register(
                'get_metrics', callable=lambda: asdict(self.metrics))
            # SharedStrategyManager.register(
            #     'get_time', callable=lambda:  pd.Timestamp(self.BROKER.get_current_time, tz='UTC').timestamp())
            
            SharedStrategyManager.register(
                'get_time', callable=lambda:  (pd.Timestamp(self.BROKER.get_current_time, tz='UTC')-datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)).total_seconds())

            self.SSM = SharedStrategyManager(
                address=('', 50000), authkey=os.getenv('SSM_PASSWORD').encode())
            print('UI Shared Memory Server configured')

        except Exception as e:
            print(f'Error in _startUISharedMemory: {e}')
            pass

    def _insightListener(self):
        """ Listen to the insights and manage the orders. """
        assert callable(
            self.executeInsight), 'executeInsight must be a callable function'
        print('Running Insight Listener')
        while self._Running:
            # TODO: could use a thread pool executor to run the executeInsight() functions
            for i in list(self.INSIGHTS):
                insight = self.INSIGHTS[i]
                if insight is None:
                    continue
                try:
                    # if self.VERBOSE > 0:
                    # print(f'Execute Insight: {
                    #     symbol}- {datetime.datetime.now()}')
                    # start_time = timeit.default_timer()

                    # Execute the insight Executors
                    passed = True
                    for executor in self.INSIGHT_EXECUTORS[insight.state]:
                        result = executor.run(
                            self.INSIGHTS[insight.INSIGHT_ID])
                        # Executor manage the insight state and mutates the insight
                        if not result.success:
                            print(f'Executor {result.executor}: {
                                  result.message}')
                            passed = False
                            break
                        elif not result.passed:
                            passed = False
                            break
                        if self.VERBOSE > 1:
                            print(
                                f'Executor {result.executor}: {result.message}')

                    if not passed:
                        continue

                    # latestInsight = self.INSIGHTS.get(insight.INSIGHT_ID)
                    # if latestInsight is not None:
                    #     self.executeInsight(self.INSIGHTS[insight.INSIGHT_ID])

                    self.executeInsight(insight)

                    # if self.VERBOSE > 0:
                    # print('Time taken executeInsight:', symbol,
                    #       timeit.default_timer() - start_time)
                except KeyError as e:
                    continue
                except Exception as e:
                    print('Error in _insightListener:', e)
                    continue
            if self.MODE == IStrategyMode.BACKTEST:
                try:
                    self.BROKER.BACKTEST_FlOW_CONTROL_BARRIER.wait()
                except BrokenBarrierError as e:
                    # print('Error in _insightListener:', e)
                    # TODO: Should be checking if the backtesting range is completed but for now we will just break
                    break
            else:
                # await asyncio.sleep(1)
                sleep(self.insightRateLimit)
            # Update the account and positions
            self.ACCOUNT = self.BROKER.get_account()
            self.POSITIONS = self.BROKER.get_positions()
            if self.POSITIONS == None:
                self.POSITIONS = {}
        print("End of Insight Listener")

    async def _on_trade_update(self, trade):
        """ format the trade stream to the strategy. """
        orderdata, event = self.BROKER.format_on_trade_update(trade)
        # check if there is data and that the order symbol is in the universe
        if orderdata and orderdata['asset']['symbol'] not in self.UNIVERSE:
            # 'Order not in universe'
            return
        print(
            f"Order: {event:<16} {orderdata['created_at']}: {orderdata['asset']['symbol']:^6}: {str(orderdata['filled_qty']):^8} / {orderdata['qty']:^8} : {orderdata['type']} / {orderdata['order_class']} : {orderdata['side']} @ {orderdata['limit_price'] if orderdata['limit_price'] != None else orderdata['filled_price']} - {orderdata['order_id']}")
        self.ORDERS[orderdata["order_id"]] = orderdata
        for i, insight in self.INSIGHTS.items():
            # Check if the insight is managing the order by checking the symbol
            if insight.symbol != orderdata['asset']['symbol']:
                continue

            match insight.state:
                case InsightState.EXECUTED:
                    # We aleady know that the order has been executed because it will never be in the insights list as executed if it was not accepted by the broker
                    if insight.order_id == orderdata['order_id']:
                        match event:
                            # case ITradeUpdateEvent.PENDING_NEW:
                            case ITradeUpdateEvent.NEW | ITradeUpdateEvent.PENDING_NEW:
                                if orderdata['legs']:
                                    insight.updateLegs(
                                        legs=orderdata['legs'])
                                return

                            case ITradeUpdateEvent.FILLED:
                                if orderdata['legs']:
                                    insight.updateLegs(
                                        legs=orderdata['legs'])
                                try:

                                    # Update the insight with the filled price
                                    insight.positionFilled(
                                        orderdata['filled_price'] if orderdata['filled_price'] != None else orderdata['limit_price'], orderdata["filled_qty"], orderdata['order_id'])
                                    self.MATRICS.positionOpened()
                                except AssertionError as e:
                                    continue
                                if insight.PARANT == None and len(insight.CHILDREN) > 0:
                                    # set the childe insight to be active
                                    for uid, childInsight in insight.CHILDREN.items():
                                        self.add_insight(childInsight)
                                    pass
                                return

                            case ITradeUpdateEvent.PARTIAL_FILLED:
                                # keep track of partial fills as some positions may be partially filled and not fully filled. in these cases we need to update the insight with the filled quantity and price
                                insight.partialFilled(
                                    orderdata['filled_qty'])
                                return

                            case ITradeUpdateEvent.CANCELED:
                                # check if we have been partially filled and remove the filled quantity from the insight
                                if insight._partial_filled_quantity != None and self.shouldClosePartialFilledIfCancelled:
                                    insight.updateState(
                                        InsightState.FILLED, 'Order Canceled, Closing Partial Filled Position')
                                    oldQuantity = insight.quantity
                                    if insight.uses_contract_size:
                                        insight.update_contracts(
                                            insight._partial_filled_quantity)
                                    else:
                                        insight.update_quantity(
                                            insight._partial_filled_quantity)
                                    if insight.close():
                                        pass
                                    else:
                                        print("Partial Filled Quantity Before Canceled: ",
                                              insight._partial_filled_quantity, " / ", oldQuantity, " - And Failed to close the position")
                                else:
                                    insight.updateState(
                                        InsightState.CANCELED, 'Order Canceled')
                                return

                            case  ITradeUpdateEvent.REJECTED:
                                insight.updateState(
                                    InsightState.REJECTED, 'Order Rejected')
                                return
                            case _:
                                pass
                case InsightState.FILLED | InsightState.CLOSED:
                    # Check if the position has been closed via SL or TP
                    # if insight.state == InsightState.CANCELED and insight._partial_filled_quantity == None:
                    #     # Check if we has a partial fill and need to get the results of the partial fill that was closed
                    #     break
                    if ((event == ITradeUpdateEvent.FILLED) or (event == ITradeUpdateEvent.CLOSED)):

                        if insight.state == InsightState.CLOSED:
                            # Insight has already been closed
                            continue
                        # check partials closes before closing the position
                        if len(insight.partial_closes) > 0:
                            for p, partialClose in enumerate(insight.partial_closes):
                                if partialClose['order_id'] == orderdata['order_id']:
                                    if (event == ITradeUpdateEvent.FILLED) or (event == ITradeUpdateEvent.CLOSED):
                                        # Update the insight closed price
                                        if self.MODE != IStrategyMode.BACKTEST:
                                            insight.partial_closes[p].set_filled_price(
                                                orderdata['filled_price'] if orderdata['filled_price'] != None else orderdata['limit_price'])
                                        else:
                                            insight.partial_closes[p].set_filled_price(
                                                orderdata['stop_price'] if orderdata['stop_price'] != None else orderdata['filled_price'])

                                        closePnL = insight.partial_closes[p].getPL(
                                        )
                                        if insight.uses_contract_size:
                                            # TODO: might not to do this if quantity is in units of the contract size
                                            closePnL = closePnL * \
                                                insight.ASSET.get(
                                                    "contract_size")
                                        # Update the metrics with the closed PnL
                                        self.MATRICS.positionClosed(closePnL)
                                        return
                        # Make sure the order is part of the insight as we dont have a clear way to tell if the closed fill is part of the strategy- to ensure that the the strategy is managed well
                        if (((orderdata['qty'] == insight.quantity) and (orderdata['side'] != insight.side) and insight.close_order_id == None)) or \
                            ((insight.close_order_id != None) and (insight.close_order_id == orderdata['order_id'])) or \
                            (insight.order_id == orderdata['order_id']) or \
                            (insight.legs != None and
                                ((insight.takeProfitOrderLeg != None and orderdata['order_id'] == insight.takeProfitOrderLeg['order_id']) or
                                    (insight.stopLossOrderLeg != None and orderdata['order_id'] == insight.stopLossOrderLeg['order_id']) or
                                    (insight.trailingStopOrderLeg !=
                                     None and orderdata['order_id'] == insight.trailingStopOrderLeg['order_id'])
                                 )
                             ):
                            # Update the insight closed price
                            try:
                                if self.MODE != IStrategyMode.BACKTEST and self.broker.NAME != ISupportedBrokers.PAPER:
                                    insight.positionClosed(
                                        orderdata['filled_price'] if orderdata['filled_price'] != None else orderdata['limit_price'], orderdata['order_id'], orderdata['filled_qty'])
                                else:
                                    insight.positionClosed(
                                        orderdata['stop_price'] if orderdata['stop_price'] != None else orderdata['filled_price'], orderdata['order_id'], orderdata['filled_qty'])

                                closePnL = 0
                                if insight.side == IOrderSide.BUY:
                                    closePnL = round(
                                        ((insight.close_price - insight.limit_price) * insight.quantity), 2)
                                else:
                                    closePnL = round(
                                        ((insight.limit_price - insight.close_price) * insight.quantity), 2)
                                if insight.uses_contract_size:
                                    # TODO: might not to do this if quantity is in units of the contract size
                                    closePnL = closePnL * \
                                        insight.ASSET.get("contract_size")

                                self.MATRICS.positionClosed(closePnL)
                                return  # No need to continue
                            except AssertionError as e:
                                continue

                    continue

        # TODOL Check if the order is part of the resolution of the strategy and has a insight that is managing it.

    def _loadUniverse(self):
        """ Loads the universe of the strategy."""
        assert callable(self.universe), 'Universe must be a callable function'
        universeSet = set(self.universe())
        for symbol in universeSet:
            self._loadAsset(symbol)
        assert (len(self.UNIVERSE) != 0), 'No assets loaded into the universe'

        for asset in self.UNIVERSE.values():
            # Init all assets in the strategy
            self._init(asset)

    def _loadAsset(self, symbol: str):
        """ Loads the asset into the universe of the strategy."""
        symbol = symbol.upper()
        assetInfo = self.BROKER.get_ticker_info(symbol)
        if assetInfo and assetInfo['status'] == 'active' and assetInfo['tradable']:
            self.UNIVERSE[assetInfo['symbol']] = assetInfo
            self.HISTORY[assetInfo['symbol']] = pd.DataFrame()
            print(
                f'Loaded {symbol}:{assetInfo["exchange"], }  into universe')
        else:
            print(f'Failed to load {symbol} into universe')

    def add_events(self, eventType: Literal['trade', 'quote', 'bar', 'news'] = 'bar', **kwargs):
        """ Adds bar streams to the strategy."""
        match eventType:
            case 'bar':
                options: IMarketDataStream = {
                }

                if self.MODE == IStrategyMode.BACKTEST:
                    # Check if we should apply TA to the stream at the start of the backtest
                    if kwargs.get('applyTA', True):
                        options['applyTA'] = True
                        options['TA'] = self.TaStrategy
                        self.BACKTESTING_CONFIG['preemptiveTA'] = True
                    else:
                        options['applyTA'] = False
                        self.BACKTESTING_CONFIG['preemptiveTA'] = False

                    options['start'] = self.BROKER.START_DATE
                    options['end'] = self.BROKER.END_DATE
                    options['stored'] = kwargs.get('stored', False)
                    options['stored_path'] = kwargs.get('stored_path', "data")

                if kwargs.get('time_frame'):
                    options['time_frame'] = kwargs.get('time_frame')
                    if options['time_frame'] != self.RESOLUTION and not self.broker.supportedFeatures.featuredBarDataStreaming:
                        print(
                            "Feature Bar Data Streaming is not supported by the broker")
                        return
                else:
                    options['time_frame'] = self.RESOLUTION
                    options['feature'] = None

                for assetInfo in self.UNIVERSE.values():
                    assetDataStreamInfo = options.copy()
                    assetDataStreamInfo['symbol'] = assetInfo.get('symbol')
                    assetDataStreamInfo['exchange'] = assetInfo.get('exchange')
                    assetDataStreamInfo['asset_type'] = assetInfo.get(
                        'asset_type')
                    if options['time_frame'] != self.RESOLUTION:
                        assetDataStreamInfo['feature'] = f"{assetDataStreamInfo['symbol']}.{
                            assetDataStreamInfo['time_frame']}"
                        if assetDataStreamInfo['feature'] in self.HISTORY:
                            if not isinstance(self.HISTORY[assetDataStreamInfo['feature']], pd.DataFrame):
                                self.HISTORY[assetDataStreamInfo['feature']
                                             ] = pd.DataFrame()
                        else:
                            self.HISTORY[assetDataStreamInfo['feature']
                                         ] = pd.DataFrame()

                    assetDataStreamInfo['type'] = eventType
                    self.STREAMS.append(
                        IMarketDataStream(**assetDataStreamInfo))
            case _:
                print(f"{eventType} Event not supported")

    def add_alpha(self, alpha: BaseAlpha):
        """ Adds an alpha to the strategy."""
        assert isinstance(
            alpha, BaseAlpha), 'alpha must be of type BaseAlpha object'

        alpha.registerAlpha()

    def add_alphas(self, alphas: List[BaseAlpha]):
        """ Adds a list of alphas to the strategy."""
        assert isinstance(
            alphas, List), 'alphas must be of type List[BaseAlpha] object'
        for alpha in alphas:
            self.add_alpha(alpha)

    def add_executor(self, executor: BaseExecutor):
        """ Adds an executor to the strategy."""
        assert isinstance(
            executor, BaseExecutor), 'executor must be of type BaseExecutor object'

        self.INSIGHT_EXECUTORS[executor.state].append(executor)

    def add_executors(self, executors: List[BaseExecutor]):
        """ Adds a list of executors to the strategy."""
        assert isinstance(
            executors, List), 'executors must be of type List[BaseExecutor] object'
        for executor in executors:
            self.add_executor(executor)

    def add_ta(self, ta: List[dict]):
        """ Adds a technical analysis to the strategy."""
        assert isinstance(
            ta, List), 'ta must be of type List[dict] object'
        if len(ta) == 0:
            return

        for ind in ta:
            if ind not in self.TaStrategy.ta:
                self.TaStrategy.ta.append(ind)

    def add_insight(self, insight: Insight):
        """ Adds an insight to the strategy."""
        assert isinstance(
            insight, Insight), 'insight must be of type Insight object'

        insight.set_mode(self.BROKER, self.assets[insight.symbol], self.MODE)

        self.INSIGHTS[insight.INSIGHT_ID] = insight

    async def _on_bar(self, bar: Any, timeframe: ITimeFrame):
        """ format the bar stream to the strategy. """
        try:
            # set_index(['symbol', 'timestamp']
            data: pd.DataFrame = None
            if (self.MODE != IStrategyMode.BACKTEST and self.BROKER.NAME != ISupportedBrokers.PAPER) and not isinstance(bar, pd.DataFrame):
                data = self.BROKER.format_on_bar(bar)
            else:
                data = bar

            if data.empty:
                print('Bar is None')
                return

            self.ACCOUNT = self.BROKER.get_account()
            self.POSITIONS = self.BROKER.get_positions()
            self.ORDERS = self.BROKER.get_orders()

            if self.ORDERS == None:
                self.ORDERS = {}

            if self.POSITIONS == None:
                self.POSITIONS = {}

            if not data.empty:
                symbol = data.index[0][0]
                timestamp = data.index[0][1]
                if symbol == None:
                    if self.VERBOSE > 0:
                        print('Symbol is None - ignoring BaseStrategy_on_bar')
                    return
                # Check if the bar is part of the resolution of the strategy or if it is a feature event
                isFeature = False
                for stream in self.STREAMS:
                    if stream["type"] == "bar" and (stream['symbol'] == symbol or stream['feature'] == symbol) and stream['time_frame'].value != self.resolution.value:
                        if stream["time_frame"].value == timeframe.value and stream["time_frame"].is_time_increment(timestamp):
                            isFeature = True
                            # Update the feature symbol name
                            if stream['feature'] != symbol:
                                data.rename(
                                    index={symbol: stream['feature']}, inplace=True)
                                symbol = stream['feature']
                            if self.VERBOSE > 0:
                                print(f'Feature Bar is part of the resolution of the strategy: {
                                      symbol} - {timestamp} - {datetime.datetime.now()}')
                            break

                if (self.resolution.value == timeframe.value and self.resolution.is_time_increment(timestamp)) or isFeature:
                    if self.VERBOSE > 0:
                        print(f'New Bar is part of the resolution of the strategy: {
                              symbol} - {timestamp} - {datetime.datetime.now()}')
                        start_time = timeit.default_timer()

                    # TODO: check if the broker sends out multiple bar data when streaming bars with the same symbol

                    self.HISTORY[symbol] = pd.concat(
                        [self.HISTORY[symbol], data])
                    # Remove duplicates keys in the history as sometimes when getting warm up data we get duplicates
                    self.HISTORY[symbol] = self.HISTORY[symbol].loc[~self.HISTORY[symbol].index.duplicated(
                        keep='first')]
                    # Needs to be warm up
                    if (len(self.HISTORY[symbol]) < self.WARM_UP):
                        print(f"Waiting for warm up: {
                              symbol} - {len(self.HISTORY[symbol])} / {self.WARM_UP}")
                        return

                    if (not self.BACKTESTING_CONFIG.get('preemptiveTA') and self.MODE == IStrategyMode.BACKTEST) or (self.MODE != IStrategyMode.BACKTEST):
                        # Run the pandas TA
                        self.HISTORY[symbol].ta.strategy(self.TaStrategy)

                    # Call the on_bar function and process the bar
                    try:
                        if (isFeature == False) or (isFeature == True and self.tradeOnFeatureEvents):
                            self.on_bar(symbol, data)
                            self._generateInsights(symbol)
                    except Exception as e:
                        print('Error in on_bar:', e)

                    if self.VERBOSE > 0:
                        print('Time taken on_bar:', symbol,
                              timeit.default_timer() - start_time)
                else:
                    """ Bar is not part of the resolution of the strategy  nor is it a feature event"""
                    pass
                    # Will need to take the the 1 min bar and convert it to the resolution of the strategy
            else:
                print('Data is empty - BROKER.format_on_bar(bar) not working')
        except Exception as e:
            print('Error in _on_bar_format in base Strategy:', e)

    def submit_order(self, insight: Insight):
        """ Submits an order to the broker."""
        assert isinstance(
            insight, Insight), 'insight must be of type Insight object'
        try:
            order = self.BROKER.execute_insight_order(
                insight, self.assets[insight.symbol])
            return order
        except BaseException as e:
            # print('Error in submit_order:', e)
            raise e

    def close_position(self, symbol, qty=None, percent=None):
        """ Cancels an order to the broker."""
        return self.BROKER.close_position(symbol, qty, percent)

    @property
    def account(self) -> Optional[IAccount]:
        """ Returns the account of the strategy."""

        return self.ACCOUNT

    @property
    def positions(self) -> dict[str, IPosition]:
        """ Returns the positions of the strategy."""
        return self.POSITIONS

    @property
    def orders(self) -> Optional[dict[str, IOrder]]:
        """ Returns the orders of the strategy."""
        return self.ORDERS

    @property
    def history(self) -> dict[str, pd.DataFrame]:
        """ Returns the candle history of the strategy."""
        return self.HISTORY

    @property
    def insights(self) -> dict[UUID, Insight]:
        """ Returns the insights of the strategy."""
        return self.INSIGHTS

    @property
    def state(self) -> dict:
        """ Returns the state of the strategy."""
        return self.VARIABLES

    @state.setter
    def state(self, state: AttributeDict):
        """ Sets the state of the strategy."""
        self.VARIABLES = state

    @property
    def broker(self) -> BaseBroker:
        """ Returns the broker used by the strategy."""
        return self.BROKER

    @property
    def assets(self) -> dict[str, IAsset]:
        """ Returns the universe of the strategy."""
        return self.UNIVERSE

    @property
    def metrics(self) -> IStrategyMatrics:
        """ Returns the metrics of the strategy."""
        return self.MATRICS

    @property
    def resolution(self) -> ITimeFrame:
        """ Returns the resolution of the strategy."""
        return self.RESOLUTION

    @property
    def tools(self) -> ITradingTools:
        """ Returns the tools of the strategy."""
        return self.TOOLS

    @property
    def streams(self) -> List[IMarketDataStream]:
        """ Returns the streams of the strategy."""
        return self.STREAMS

    @property
    def warm_up(self) -> int:
        """ Returns the warm up period of the strategy."""
        return self.WARM_UP

    @warm_up.setter
    def warm_up(self, warm_up: int):
        """ Sets the warm up period of the strategy."""
        if warm_up < 0:
            raise ValueError('Warm up period must be greater than 0')
        if warm_up <= self.WARM_UP:
            # print('Warm up period cannot be reduced')
            return
        self.WARM_UP = warm_up




import dash
from dash import html, Output, Input, callback
import numpy as np


def AccountBalanceCard(balance: float, alltimeChange: float, dailyChange: float):
    currencySymbol = "£"
    return html.Div([
        html.H3("Account Balance",
                className="text-xl font-semibold text-accent mb-2"),
        html.P(f"{currencySymbol}{balance}",
               className="text-2xl font-bold text-white"),
        html.P(f"{currencySymbol}{alltimeChange} ~ {np.round((((balance+alltimeChange)/balance)-1)*100, 2)
                                                    }%", className=f"text-sm {'text-green-500' if alltimeChange > 0 else 'text-red-500' if alltimeChange < 0 else 'text-wite'}"),
        html.P(f"{currencySymbol}{dailyChange} ~ {np.round((((balance+dailyChange)/balance)-1)*100, 2)
                                                  }% (24h)", className=f"text-sm {'text-green-500' if dailyChange > 0 else 'text-red-500' if dailyChange < 0 else 'text-wite'}"),
    ], 
    className="bg-primary-foreground p-6 rounded-lg shadow-lg border border-accent")




import asyncio
from concurrent.futures import ThreadPoolExecutor
import os
import datetime
from time import sleep
from types import NoneType
import uuid
import numpy as np
from typing import Awaitable, List, Literal, Optional, overload
from collections import deque
from threading import Barrier, BrokenBarrierError
from concurrent.futures import as_completed
from pathlib import Path

import vectorbt as vbt

import pandas as pd
from tqdm import tqdm

from .base_broker import BaseBroker
from .interfaces import IQuote, ISupportedBrokerFeatures, ITimeInForce, ISupportedBrokers, IOrderClass, IOrderRequest, IOrderSide, IOrderType, ITimeInForce, ITradeUpdate, ITradeUpdateEvent
from ..insight.insight import Insight
from .interfaces import IAccount, IOrder, IPosition, IAsset, IOrderLegs, IOrderLeg
from ..strategy.interfaces import IMarketDataStream, IStrategyMode
from ..utils.timeframe import ITimeFrame, ITimeFrameUnit


import yfinance as yf


class PaperBroker(BaseBroker):
    MODE: IStrategyMode = IStrategyMode.BACKTEST

    ACCOUNT: IAccount = None
    Positions: dict[str, dict[uuid.UUID, IPosition]] = {}
    Orders: dict[uuid.UUID, IOrder] = {}
    LEVERAGE: int = 4

    # Backtest mode
    STARTING_CASH: float = 100_000.00
    START_DATE: datetime.datetime = None
    END_DATE: datetime.datetime = None
    CurrentTime: datetime.datetime = None
    PreviousTime: datetime.datetime = None
    HISTORICAL_DATA: dict[str, dict[Literal['trade',
                                            'quote', 'bar', 'news', 'signals'], pd.DataFrame]] = {}
    RUNNING_TRADE_STREAM: bool = False
    RUNNING_MARKET_STREAM: bool = False
    BACKTEST_FlOW_CONTROL_BARRIER: Barrier = None

    PENDING_ORDERS: deque[IOrder] = deque()
    ACTIVE_ORDERS: deque[IOrder] = deque()
    CLOSE_ORDERS: deque[IOrder] = deque()
    CANCELED_ORDERS: deque[IOrder] = deque()

    ACCOUNT_HISTORY: dict[datetime.date, IAccount] = {}

    _MARKET_STREAMS: dict[IMarketDataStream, asyncio.Future] = {}
    """Market Streams"""

    FeedDelay: int = 0
    # DEBUG
    VERBOSE: int = 0

    def __init__(self, cash: float = 100_000.00, start_date: datetime.date = None, end_date: datetime.date = None, leverage: int = 4, currency: str = "GBP", allow_short: bool = True, mode: IStrategyMode = IStrategyMode.BACKTEST, feed: Literal['yf', 'eod'] = 'yf', feedDelay: int = 0, verbose: int = 0):

        super().__init__(ISupportedBrokers.PAPER, True, feed)
        self.MODE = mode
        self.VERBOSE = verbose
        self.LEVERAGE = leverage
        self.STARTING_CASH = cash
        self.ACCOUNT = IAccount(account_id="PAPER_ACCOUNT", equity=self.STARTING_CASH, cash=self.STARTING_CASH, currency=currency,
                                buying_power=cash*self.LEVERAGE, leverage=self.LEVERAGE, shorting_enabled=allow_short)

        # Set the backtest configuration
        if self.MODE == IStrategyMode.BACKTEST:
            assert start_date and end_date, 'Start and End date must be provided for backtesting'
            assert start_date < end_date, 'Start date must be before end date'
            # self.START_DATE = start_date.replace(tzinfo=datetime.timezone.utc)
            self.START_DATE = start_date
            self.END_DATE = end_date
            self.CurrentTime = self.START_DATE
            self.update_account_history()
            self.BACKTEST_FlOW_CONTROL_BARRIER = Barrier(3)
            # self.BACKTEST_FlOW_CONTROL_BARRIER.reset()
        else:
            print("Live Papaer Trading Mode - There is a:",
                  feedDelay, "minute delay in the feed")
            self.FeedDelay = feedDelay
            self.CurrentTime = datetime.datetime.now(
            ) - datetime.timedelta(minutes=self.FeedDelay)

        self.supportedFeatures = ISupportedBrokerFeatures(
            barDataStreaming=True, featuredBarDataStreaming=True, trailingStop=False)

    def get_ticker_info(self, symbol: str):
        if symbol in self.TICKER_INFO:
            return self.TICKER_INFO[symbol]

        if self.DataFeed == 'yf':
            symbol = symbol.replace('/', '-')
            try:
                yfRes = yf.Ticker(symbol)
                if not yfRes:
                    return None
                tickerInfo = yfRes.info
                tickerAsset: IAsset = IAsset(
                    id=tickerInfo['uuid'],
                    name=tickerInfo['shortName'],
                    asset_type=tickerInfo["quoteType"],
                    exchange=tickerInfo["exchange"],
                    symbol=tickerInfo["symbol"],
                    status="active",
                    tradable=True,
                    marginable=True,
                    shortable=self.Account.shorting_enabled,
                    fractionable=True,
                    min_order_size=0.001,
                    min_price_increment=1 /
                    np.power(10, tickerInfo["priceHint"]
                             ) if "priceHint" in tickerInfo else 0.01
                )
                self.TICKER_INFO[symbol] = tickerAsset
                return tickerAsset
            except Exception as e:
                print("Error: ", e)
                return None
        else:
            raise NotImplementedError(
                f'DataFeed {self.DataFeed} not supported')

    def get_history(self, asset: IAsset, start: datetime.datetime, end: datetime.datetime, resolution: ITimeFrame, shouldDelta: bool = True) -> pd.DataFrame:
        super().get_history(asset, start, end, resolution)

        if self.DataFeed == 'yf':
            symbol = asset['symbol'].replace('/', '-')
            formatTF = f'{resolution.amount_value}{
                resolution.unit_value[0].lower()}'
            if self.MODE == IStrategyMode.BACKTEST:
                delta: datetime.timedelta = start - \
                    self.get_current_time if shouldDelta else datetime.timedelta()
                # print("start: ", self.get_current_time-start, "end: ", self.get_current_time-end)
                data = yf.download(
                    symbol, start=resolution.get_time_increment(start-delta), end=resolution.get_time_increment(end-delta), interval=formatTF, progress=False)
            else:
                data = yf.download(
                    symbol, start=start, end=end, interval=formatTF, progress=False)

            return self.format_on_bar(data, asset['symbol'])
        else:
            raise NotImplementedError(
                f'DataFeed {self.DataFeed} not supported')

    def get_account(self):
        return self.Account

    def get_position(self, symbol: str, mode: Literal["all", "active"] = 'active'):
        positions = self.Positions.get(symbol)
        if positions:
            aggregated_position = {}
            for uid, position in positions.items():
                if position['qty'] != 0 and position['asset']['symbol'] == symbol:
                    if not aggregated_position:
                        aggregated_position = position.copy()
                    else:
                        aggregated_position['qty'] += position['qty']
                        aggregated_position['market_value'] += position['market_value']
                        aggregated_position['unrealized_pl'] += position['unrealized_pl']
                        aggregated_position['cost_basis'] += position['cost_basis']

            if (aggregated_position == None or (aggregated_position.get('qty') == 0 or not aggregated_position.get('qty')) and mode == 'active'):
                # del self.Positions[symbol]
                return None
            else:
                # Update the side
                aggregated_position['side'] = IOrderSide.BUY if aggregated_position['qty'] > 0 else IOrderSide.SELL

            return aggregated_position

    def get_positions(self, mode: Literal["all", "active"] = 'active'):
        aggregated_positions = {}
        for symbol in self.Positions.keys():
            aggregated_positions[symbol] = self.get_position(symbol)

        return aggregated_positions

    def get_orders(self):
        # active orders
        returned_orders: dict[str, IOrder] = {}
        active_orders = [order for order in self.Orders.values(
        ) if order['status'] == ITradeUpdateEvent.NEW or order['status'] == ITradeUpdateEvent.FILLED]
        if len(active_orders) == 0:
            return None
        for order in active_orders:
            returned_orders[order['order_id']] = order

        return returned_orders

    def get_order(self, order_id):
        return self.Orders.get(order_id)

    def format_order(self, order) -> IOrder:
        assert isinstance(order, IOrder), "Order must be a dictionary"
        return order

    def cancel_order(self, order_id: str):
        order = self.Orders.get(order_id)
        if order:
            if order['status'] == ITradeUpdateEvent.FILLED:
                raise BaseException({
                    "code": "already_filled",
                    "data": {"order_id": order_id}
                })
            elif order['status'] == ITradeUpdateEvent.CANCELED or order in self.CANCELED_ORDERS:
                raise BaseException({
                    "code": "already_canceled",
                    "data": {"order_id": order_id}
                })
            else:
                order['updated_at'] = self.get_current_time
                self.CANCELED_ORDERS.append(order)
                return order
        else:
            # check legs
            for i, leg_order in self.Orders.items():
                if leg_order['legs'] != None and leg_order not in self.CANCELED_ORDERS and (leg_order['status'] != ITradeUpdateEvent.CANCELED and leg_order['status'] != ITradeUpdateEvent.CLOSED):
                    if leg_order['legs'].get('take_profit'):
                        if leg_order['legs']['take_profit']['order_id'] == order_id:
                            # remove the take profit leg
                            del self.Orders[i]['legs']['take_profit']
                            return order_id

                    if leg_order['legs'].get('stop_loss'):
                        if leg_order['legs']['stop_loss']['order_id'] == order_id:
                            del self.Orders[i]['legs']['stop_loss']
                            return order_id

        # Order Id not found
        raise BaseException({
            "code": "order_not_found",
            "data": {"order_id": order_id}
        })

    def get_latest_quote(self, asset: IAsset) -> IQuote:
        # FIXME: Temp: Get the latest quote at bar close and bid ask spread at is high and low
        if self.DataFeed == 'yf':
            currentBar = self._get_current_bar(
                asset['symbol']).iloc[0]
            quote = self.format_on_quote(currentBar)
            quote['symbol'] = asset['symbol']
            return quote

        else:
            raise NotImplementedError(
                f'DataFeed {self.DataFeed} not supported')

    def startTradeStream(self, callback):
        super().startTradeStream(callback)
        self.RUNNING_TRADE_STREAM = True
        loop = asyncio.new_event_loop()
        if self.MODE == IStrategyMode.BACKTEST:
            # trade stream for all of the pending, filled, canceled oerders.
            while self.get_current_time <= self.END_DATE and self.RUNNING_TRADE_STREAM:
                try:
                    self.BACKTEST_FlOW_CONTROL_BARRIER.wait()

                    self.processPendingOrders(callback, loop)
                    self.processCanceledOrders(callback, loop)
                    self.processActiveOrders(callback, loop)
                    self.processClosedOrders(callback, loop)

                except BrokenBarrierError:
                    continue
                except Exception as e:
                    print("Error: ", e)
                    continue
            print("End of Trade Stream")

        else:
            # live paper trade stream
            try:
                while self.RUNNING_TRADE_STREAM:
                    self.processPendingOrders(callback, loop)
                    self.processCanceledOrders(callback, loop)
                    self.processActiveOrders(callback, loop)
                    self.processClosedOrders(callback, loop)
                    sleep(1)
            except Exception as e:
                print("Error: ", e)

    def processPendingOrders(self, callback: Awaitable, loop: asyncio.AbstractEventLoop):
        for i, order in enumerate(list(self.PENDING_ORDERS)):
            currentBar = self._get_current_bar(
                order['asset']['symbol'])
            if currentBar is None:
                continue
            currentBar = currentBar.iloc[0]

            if order['created_at'] == self.get_current_time:
                order['status'] = ITradeUpdateEvent.NEW
                self._update_order(order)
                loop.run_until_complete(
                    callback(ITradeUpdate(order, order['status'])))

            if order['type'] == IOrderType.MARKET:
                # Market order - FILLED at the current close price
                order['filled_price'] = currentBar.open
                order['filled_qty'] = order['qty']
                order['status'] = ITradeUpdateEvent.FILLED
                order['filled_at'] = self.get_current_time
                order['updated_at'] = self.get_current_time

                try:
                    self._update_order(order)
                    loop.run_until_complete(callback(ITradeUpdate(
                        order, order['status'])))
                except BaseException as e:
                    if e.code == "insufficient_funds":
                        print("Error: ", e)

                continue

            elif order['type'] == IOrderType.LIMIT:
                """
                sometime the currentbar high and low can be the same as the close price when using minute data with yfinance API
                """
                autopass = False
                if currentBar.high == currentBar.low:
                    pricePadding = 0.001  # 0.1%
                    if ((currentBar.high*(1+pricePadding)) >= order['limit_price'] >= (currentBar.low*(1-pricePadding))):
                        autopass = True

                # we need to check if the limit price is within the high and low of the current bar
                if (order['limit_price'] >= currentBar.low and order['limit_price'] <= currentBar.high) or autopass:
                    order['filled_price'] = order['limit_price']
                    order['filled_qty'] = order['qty']
                    order['status'] = ITradeUpdateEvent.FILLED
                    order['filled_at'] = self.get_current_time
                    order['updated_at'] = self.get_current_time
                    try:
                        self._update_order(order)
                        loop.run_until_complete(callback(ITradeUpdate(
                            order, order['status'])))
                    except BaseException as e:
                        if e.code == "insufficient_funds":
                            print("Error: ", e)
                    continue

    def processActiveOrders(self, callback: Awaitable, loop: asyncio.AbstractEventLoop):
        for i, order in enumerate(list(self.ACTIVE_ORDERS)):
            # update the position information as the position is filled and keep track of all  positions PNL
            self._update_position(order)

            # check if the order has take profit or stop loss
            if order['legs']:
                currentBar = self._get_current_bar(
                    order['asset']['symbol'])
                if currentBar is None:
                    continue
                currentBar = currentBar.iloc[0]
                # FIXME: Figure out if the take profit or stop loss is hit first

                if order['legs'].get('take_profit'):
                    take_profit = order['legs']['take_profit']
                    if (currentBar.high >= take_profit['limit_price'] >= currentBar.low) or ((currentBar.close >= take_profit['limit_price'] and order['side'] == IOrderSide.BUY) or (currentBar.close <= take_profit['limit_price'] and order['side'] == IOrderSide.SELL)):
                        take_profit['filled_price'] = take_profit['limit_price']
                        take_profit['status'] = ITradeUpdateEvent.CLOSED
                        take_profit['filled_at'] = self.get_current_time
                        take_profit['updated_at'] = self.get_current_time

                        order['stop_price'] = take_profit['limit_price']
                        order['updated_at'] = self.get_current_time
                        order['status'] = take_profit['status']
                        order['legs']['take_profit'] = take_profit
                        order['side'] = IOrderSide.SELL if order['side'] == IOrderSide.BUY else IOrderSide.BUY

                        self._update_order(order)
                        loop.run_until_complete(callback(ITradeUpdate(
                            order, order['status'])))
                        continue
                if order['legs'].get('stop_loss'):
                    stop_loss = order['legs']['stop_loss']
                    if (currentBar.high >= stop_loss['limit_price'] >= currentBar.low or ((currentBar.close <= stop_loss['limit_price'] and order['side'] == IOrderSide.BUY) or (currentBar.close >= stop_loss['limit_price'] and order['side'] == IOrderSide.SELL))):
                        stop_loss['filled_price'] = stop_loss['limit_price']
                        stop_loss['status'] = ITradeUpdateEvent.CLOSED
                        stop_loss['filled_at'] = self.get_current_time
                        stop_loss['updated_at'] = self.get_current_time

                        order['stop_price'] = stop_loss['limit_price']
                        order['updated_at'] = self.get_current_time
                        order['status'] = stop_loss['status']
                        order['legs']['stop_loss'] = stop_loss
                        order['side'] = IOrderSide.SELL if order['side'] == IOrderSide.BUY else IOrderSide.BUY

                        self._update_order(order)
                        loop.run_until_complete(callback(ITradeUpdate(
                            order,  order['status'])))
                        continue
            else:
                # USually a market order or limit order without take profit or stop loss
                continue

    def processClosedOrders(self, callback: Awaitable, loop: asyncio.AbstractEventLoop):
        for i, order in enumerate(list(self.CLOSE_ORDERS)):
            currentBar = self._get_current_bar(
                order['asset']['symbol'])
            if currentBar is None:
                continue
            currentBar = currentBar.iloc[0]

            order['stop_price'] = currentBar.open
            order['filled_qty'] = order['qty']
            order['status'] = ITradeUpdateEvent.CLOSED
            order['filled_at'] = self.get_current_time
            order['updated_at'] = self.get_current_time
            self._update_order(order)

            loop.run_until_complete(callback(ITradeUpdate(
                order, order['status'])))

    def processCanceledOrders(self, callback: Awaitable, loop: asyncio.AbstractEventLoop):
        for i, order in enumerate(list(self.CANCELED_ORDERS)):
            if order['status'] != ITradeUpdateEvent.FILLED and order['status'] != ITradeUpdateEvent.CLOSED:
                order['status'] = ITradeUpdateEvent.CANCELED
                order['updated_at'] = self.get_current_time
                self._update_order(order)
                loop.run_until_complete(callback(ITradeUpdate(
                    order, order['status'])))
            else:
                self.CANCELED_ORDERS.remove(order)

    async def closeTradeStream(self):
        self.RUNNING_TRADE_STREAM = False
        # if self.MODE == IStrategyMode.BACKTEST:
        # else:
        #     raise NotImplementedError(f'Mode {self.MODE} not supported')

    def _update_position(self, order: IOrder):
        symbol = order['asset']['symbol']
        orderId = order['order_id']
        currentBar = self._get_current_bar(symbol)
        if currentBar is None:
            return
        currentBar = currentBar.iloc[0]
        oldPosition: IPosition = None

        # check if there is a position for the symbol
        if not self.Positions.get(symbol):
            self.Positions[symbol] = {}

        # Check if this is a tracked open position
        if self.Positions[symbol].get(orderId):
            oldPosition = self.Positions[symbol].get(orderId).copy()
            match order['status']:
                case ITradeUpdateEvent.FILLED:
                    # watch changes in the position - order filled
                    pass

                case ITradeUpdateEvent.CLOSED:
                    if order['stop_price']:
                        if self.Positions[symbol][orderId]['side'] == IOrderSide.BUY:
                            # if order['side'] == IOrderSide.BUY:
                            assert order['side'] == IOrderSide.SELL, "Order side must be SELL for closing a BUY position"
                            self.Positions[symbol][orderId]['qty'] -= order['qty']
                        elif self.Positions[symbol][orderId]['side'] == IOrderSide.SELL:
                            assert order['side'] == IOrderSide.BUY, "Order side must be BUY for closing a SELL position"
                            self.Positions[symbol][orderId]['qty'] += order['qty']

                        # Clear buying power wwithheld by the order
                        entryPrince = order['filled_price'] if order[
                            'filled_price'] != None else self.Positions[symbol][orderId]['avg_entry_price']
                        self.Account.cash += np.round(
                            (order['qty'] * entryPrince) / self.LEVERAGE, 2)

                    else:
                        print("Order Close Without stop_price:", order)

        else:
            match order['status']:
                case ITradeUpdateEvent.FILLED:
                    # add positions dictionary
                    entryPrince = order['filled_price'] if order['filled_price'] != None else currentBar.close
                    market_value = entryPrince * order['qty']
                    self.Positions[symbol][orderId] = IPosition(
                        asset=order['asset'],
                        avg_entry_price=entryPrince,
                        qty=order['qty'] if order['side'] == IOrderSide.BUY else -order['qty'],
                        side=order['side'],
                        market_value=market_value,
                        cost_basis=market_value,
                        current_price=entryPrince,
                        unrealized_pl=0
                    )
                    marginRequired = order['qty'] * order['filled_price']
                    if self.Account.cash < marginRequired/self.LEVERAGE:
                        # Not enough buying power
                        order['status'] = ITradeUpdateEvent.CANCELED
                        self.CANCELED_ORDERS.append(order)
                        raise BaseException({
                            "code": "insufficient_funds",
                            "data": {"order_id": order['order_id']}
                        })
                    self.Account.cash -= np.round(
                        marginRequired/self.LEVERAGE, 2)
                    pass
                case ITradeUpdateEvent.CANCELED:
                    # ORder will just be canclled
                    pass

        if oldPosition:
            self.Positions[symbol][orderId]['current_price'] = currentBar.close if not order['stop_price'] else order['stop_price']

            if self.Positions[symbol][orderId]['current_price'] == oldPosition['current_price'] and order['status'] != ITradeUpdateEvent.CLOSED:
                # No price change in the position since the last update
                return

            # or order['status'] != ITradeUpdateEvent.CLOSED or not order['stop_price']:
            if oldPosition != 0:
                # Update the position market value
                self.Positions[symbol][orderId]['market_value'] = self.Positions[symbol][orderId]['current_price'] * \
                    np.abs(
                        # qunaity can be negative for short positions
                        oldPosition['qty'])

                # Update the unrealized PnL
                self.Positions[symbol][orderId]['unrealized_pl'] = (self.Positions[symbol][orderId]['market_value'] - self.Positions[symbol][orderId]['cost_basis']
                                                                    ) if self.Positions[symbol][orderId]['side'] == IOrderSide.BUY else (self.Positions[symbol][orderId]['cost_basis'] - self.Positions[symbol][orderId]['market_value'])

                # Update the account equity
                changeInPL = round(self.Positions[symbol][orderId]['unrealized_pl'] -
                                   oldPosition['unrealized_pl'], 2)
                self.Account.cash += changeInPL
                self.Account.equity += changeInPL

            if self.Positions[symbol][orderId]['qty'] == 0:
                # remove the position from the Positions dictionary
                del self.Positions[symbol][orderId]

        else:
            return

    def _log_signal(self, order: IOrder, signalType: Literal['entry', 'exit']):
        symbol = order['asset']['symbol']
        currentBarIndex = self._get_current_bar(symbol).index
        # Track the signals to later plot the signals, entries and exits and run against the backtester (ours and vector BT)
        if signalType == 'entry':
            # Log th entry signal
            if order['side'] == IOrderSide.BUY:
                self.HISTORICAL_DATA[symbol
                                     ]['signals'].loc[currentBarIndex, 'entries'] = True
            else:
                self.HISTORICAL_DATA[symbol
                                     ]['signals'].loc[currentBarIndex, 'short_entries'] = True
            # Record the entry price
            if order['filled_price']:
                if np.isnan(self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex, 'price'][0]):
                    self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex,
                                                                'price'] = order['filled_price']
                else:
                    self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex, 'price'] = (
                        self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex, 'price'][0] + order['filled_price']) / 2

        elif signalType == 'exit':
            # Log the exit signal
            if order['side'] == IOrderSide.SELL:
                self.HISTORICAL_DATA[symbol
                                     ]['signals'].loc[currentBarIndex, 'exits'] = True
            else:
                self.HISTORICAL_DATA[symbol
                                     ]['signals'].loc[currentBarIndex, 'short_exits'] = True
            # Record the exit price
            if order['stop_price']:
                if np.isnan(self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex, 'price'][0]):
                    self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex,
                                                                'price'] = order['stop_price']
                else:
                    self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex, 'price'] = (
                        self.HISTORICAL_DATA[symbol]['signals'].loc[currentBarIndex, 'price'][0] + order['stop_price']) / 2

        if order['side'] == IOrderSide.BUY:
            self.HISTORICAL_DATA[symbol
                                 ]['signals'].loc[currentBarIndex, 'qty'] += order['qty']
        else:
            self.HISTORICAL_DATA[symbol
                                 ]['signals'].loc[currentBarIndex, 'qty'] -= order['qty']

    def _update_order(self, order: IOrder):
        if self.Orders.get(order['order_id']):
            oldOrder = self.Orders[order['order_id']].copy()
        else:
            oldOrder = None

        def onNewOrder():
            # Add the new order to the pending orders queue
            if not oldOrder:
                self.PENDING_ORDERS.append(order)

                """ Check if the order is affecting any active orders as it might be an order to close a position and update the active orders accordingly"""
                if self.get_position(order['asset']['symbol']):
                    tempCloseOrder = order.copy()
                    conflicting = False
                    for i, activeOrder in enumerate(list(self.ACTIVE_ORDERS)):
                        if activeOrder['asset']['symbol'] == order['asset']['symbol'] and \
                                activeOrder['side'] != order['side'] and \
                                activeOrder['status'] == ITradeUpdateEvent.FILLED:

                            if order in self.PENDING_ORDERS:
                                # remove the order from the pending orders
                                self.PENDING_ORDERS.remove(order)

                            if activeOrder['qty'] == tempCloseOrder['qty']:
                                # close the position
                                self.ACTIVE_ORDERS.remove(activeOrder)

                                tempCloseOrder['order_id'] = activeOrder['order_id']
                                tempCloseOrder['filled_price'] = activeOrder['filled_price']
                                tempCloseOrder['filled_qty'] = activeOrder['filled_qty']

                                self.CLOSE_ORDERS.append(tempCloseOrder.copy())
                                tempCloseOrder['qty'] = 0
                                conflicting = True

                            elif (activeOrder['qty'] - tempCloseOrder['qty']) > 0:
                                # partially close the position
                                self.ACTIVE_ORDERS.remove(activeOrder)
                                # reduce the quantity of the active order
                                activeOrder['qty'] -= tempCloseOrder['qty']
                                self.ACTIVE_ORDERS.append(activeOrder)

                                # send the close order to the close orders
                                tempCloseOrder['order_id'] = activeOrder['order_id']
                                tempCloseOrder['filled_price'] = activeOrder['filled_price']
                                tempCloseOrder['filled_qty'] = tempCloseOrder['qty']
                                self.CLOSE_ORDERS.append(tempCloseOrder.copy())
                                tempCloseOrder['qty'] = 0
                                conflicting = True
                            else:
                                # close multiple positions. quantityLeft > 0
                                quantityLeft = tempCloseOrder['qty'] - \
                                    activeOrder['qty']
                                self.ACTIVE_ORDERS.remove(activeOrder)
                                # send the close order for the active order
                                tempCloseOrder['order_id'] = activeOrder['order_id']
                                tempCloseOrder['qty'] = activeOrder['qty']
                                tempCloseOrder['filled_price'] = activeOrder['filled_price']
                                tempCloseOrder['filled_qty'] = tempCloseOrder['qty']
                                self.CLOSE_ORDERS.append(tempCloseOrder.copy())

                                tempCloseOrder['qty'] = quantityLeft
                                conflicting = True
                                continue

                            if tempCloseOrder['qty'] == 0:
                                break
                    if tempCloseOrder['qty'] > 0 and conflicting:
                        # Add the order to the pending orders
                        order['qty'] = tempCloseOrder['qty']
                        self.PENDING_ORDERS.append(order)

        def onFilledOrder():
            if oldOrder in self.PENDING_ORDERS:
                self.PENDING_ORDERS.remove(oldOrder)

                # update position of the order

                self._update_position(order)
                # Add the order to the active orders
                if self.Positions[order['asset']['symbol']][order['order_id']]['qty'] != 0:
                    # check if the order has legs and update the states of the legs
                    if order['legs']:
                        if order['legs'].get('take_profit'):
                            order['legs']['take_profit']['status'] = ITradeUpdateEvent.PENDING_NEW
                            order['legs']['take_profit']['updated_at'] = self.get_current_time
                        if order['legs'].get('stop_loss'):
                            order['legs']['stop_loss']['status'] = ITradeUpdateEvent.PENDING_NEW
                            order['legs']['stop_loss']['updated_at'] = self.get_current_time
                    # Add the order to the active orders
                    self.ACTIVE_ORDERS.append(order)
                    if self.MODE == IStrategyMode.BACKTEST:
                        # log the entry signal
                        self._log_signal(order, 'entry')

            elif oldOrder in self.ACTIVE_ORDERS:
                # update position of the order
                self._update_position(order)

        def onClosedOrder():
            # This update is already done in the on trade update
            self._update_position(order)

            # Check if the position is completely closed and remove it from the positions dictionary - if the qty is 0 or None
            if self.Positions[order['asset']['symbol']].get(order['order_id']) == None:
                pass

            if oldOrder in self.ACTIVE_ORDERS:
                self.ACTIVE_ORDERS.remove(oldOrder)

            if oldOrder in self.CANCELED_ORDERS:
                self.CANCELED_ORDERS.remove(oldOrder)

            if oldOrder in self.CLOSE_ORDERS:
                self.CLOSE_ORDERS.remove(oldOrder)
            if order in self.CLOSE_ORDERS:
                # This is when the order was sent and we are closing a position with another order id than the original order id - order is the updated order with the new order id and oldOrder is the original order (in the filled state)
                self.CLOSE_ORDERS.remove(order)

            if self.MODE == IStrategyMode.BACKTEST:
                # log the exit signal
                self._log_signal(order, 'exit')

        def onCanceledOrder():
            if oldOrder in self.ACTIVE_ORDERS:
                """ Should not really happen as the order state check should be before have already happened when cancelling the order """
                # if order is already filled oldOrder['status'] == ITradeUpdateEvent.CANCELED and oldOrder in self.CANCELED_ORDERS:
                if oldOrder in self.CANCELED_ORDERS:
                    self.CANCELED_ORDERS.remove(oldOrder)

                raise BaseException({
                    "code": "already_filled",
                    "data": {"symbol": order['asset']['symbol']}
                })

            self._update_position(order)
            if (oldOrder['status'] == ITradeUpdateEvent.NEW) or (oldOrder in self.PENDING_ORDERS):
                # Remove the order from the pending orders if it is not filled and give your money back
                self.PENDING_ORDERS.remove(oldOrder)

            if oldOrder in self.CLOSE_ORDERS:
                self.CLOSE_ORDERS.remove(oldOrder)

            if oldOrder in self.CANCELED_ORDERS:
                self.CANCELED_ORDERS.remove(oldOrder)

        match order['status']:
            case ITradeUpdateEvent.NEW:
                onNewOrder()
            case ITradeUpdateEvent.FILLED:
                onFilledOrder()
            case ITradeUpdateEvent.CANCELED:
                onCanceledOrder()
            case ITradeUpdateEvent.CLOSED:
                onClosedOrder()
            case _:
                pass

        self.Orders[order['order_id']] = order
        return order

    def execute_insight_order(self, insight: Insight, asset: IAsset):
        super().execute_insight_order(insight, asset)
        orderRequest: IOrderRequest = {
            "symbol": insight.symbol,
            "qty": insight.quantity,
            "side": insight.side,
            "time_in_force": ITimeInForce.GTC,
            "order_class": IOrderClass.SIMPLE
        }

        assert insight.quantity > 0, "Order quantity must be greater than 0"

        if insight.TP and insight.SL:
            orderRequest["order_class"] = IOrderClass.BRACKET
            orderRequest["take_profit"] = insight.TP[-1]
            orderRequest["stop_loss"] = insight.SL
        elif insight.TP:
            orderRequest["order_class"] = IOrderClass.OTO
            orderRequest["take_profit"] = insight.TP[-1]
        elif insight.SL:
            orderRequest["order_class"] = IOrderClass.OTO
            orderRequest["stop_loss"] = insight.SL
        if insight.limit_price:
            orderRequest["limit_price"] = insight.limit_price

        if insight.type in IOrderType:
            orderRequest["type"] = insight.type
        else:
            print(f"Order Type not supported {insight.type}")
            return
        try:
            if orderRequest:
                # submit the new order to be executed in the next tick
                order = self._submit_order(orderRequest)
                return order

        except BaseException as e:
            raise e

    def _submit_order(self, orderRequest: IOrderRequest) -> IOrder:
        # check if the buying power is enough to place the order
        try: 
            if orderRequest['type'] == IOrderType.MARKET:
                currentBar = self._get_current_bar(
                    orderRequest['symbol'])
                if currentBar.empty:
                    return
                currentBar = currentBar.iloc[0]
                orderRequest['limit_price'] = currentBar.close

            marginRequired = orderRequest['qty'] * orderRequest['limit_price']
            buying_power = self.Account.buying_power

            # Account for the market value of the position if the order is in the opposite direction
            position_agg = self.get_position(orderRequest['symbol'])
            if position_agg:
                if (orderRequest['side'] == IOrderSide.SELL and position_agg['qty'] < 0) or \
                        (orderRequest['side'] == IOrderSide.BUY and position_agg['qty'] > 0):
                    buying_power += position_agg['market_value']

            if buying_power < marginRequired:
                raise BaseException({
                    "code": "insufficient_balance",
                    "data": {"symbol": orderRequest['symbol'],
                            "requires": marginRequired,
                            "available": self.Account.buying_power,
                            "message": "Insufficient balance to place the order"}
                })
            # check if the orderRequest is valid
            # TP and SL should already be greater or less than the limit price depending on the side and  quantity based on Insight class logic
            if orderRequest['qty'] == None or orderRequest['qty'] <= 0:
                raise BaseException({
                    "code": "invalid_order",
                    "data": {"symbol": orderRequest['symbol'],
                            "message": "Order quantity must be greater than 0"}
                })
            if (orderRequest['type'] == IOrderType.LIMIT or orderRequest['type'] == IOrderType.STOP_LIMIT) and not orderRequest['limit_price']:
                raise BaseException({
                    "code": "invalid_order",
                    "data": {"symbol": orderRequest['symbol'],
                            "message": "Limit price must be provided for limit order"}
                })

            # Set up the order legs
            legs = IOrderLegs()
            if orderRequest.get('take_profit'):
                legs["take_profit"] = IOrderLeg(
                    order_id=uuid.uuid4(), limit_price=orderRequest['take_profit'], filled_price=None, status=ITradeUpdateEvent.NEW, filled_at=None, created_at=self.get_current_time, updated_at=self.get_current_time, submitted_at=self.get_current_time)

            if orderRequest.get('stop_loss'):
                legs["stop_loss"] = IOrderLeg(
                    order_id=uuid.uuid4(), limit_price=orderRequest['stop_loss'], filled_price=None, status=ITradeUpdateEvent.NEW, filled_at=None, created_at=self.get_current_time, updated_at=self.get_current_time, submitted_at=self.get_current_time)
            if orderRequest.get('trail_price'):
                legs["trailing_stop"] = IOrderLeg(
                    order_id=uuid.uuid4(), limit_price=orderRequest['trail_price'], filled_price=None, status=ITradeUpdateEvent.NEW, filled_at=None, created_at=self.get_current_time, updated_at=self.get_current_time, submitted_at=self.get_current_time)

            order = IOrder(
                order_id=uuid.uuid4(),
                asset=self.get_ticker_info(orderRequest['symbol']),
                limit_price=orderRequest['limit_price'] if orderRequest['limit_price'] else None,
                filled_price=None,
                stop_price=None,
                qty=orderRequest['qty'],
                filled_qty=None,
                side=orderRequest['side'],
                type=orderRequest['type'],
                time_in_force=orderRequest['time_in_force'],
                status=ITradeUpdateEvent.NEW,
                order_class=orderRequest['order_class'],
                created_at=self.get_current_time,
                updated_at=self.get_current_time,
                submitted_at=self.get_current_time,
                filled_at=None,
                legs=legs

            )

            if self.Account.cash < marginRequired/self.LEVERAGE:
                # Not enough buying power
                raise BaseException({
                    "code": "insufficient_funds",
                    "data": {"order_id": order['order_id']}
                })
            # self.Account.cash -= np.round(marginRequired/self.LEVERAGE, 2)
            self._update_order(order)
            return order
        
        except BaseException as e:
            raise  e

    def update_account_history(self):
        if self.VERBOSE >= 2:
            print("Updating Account History")
            print("Account: ", self.Account)
        self.ACCOUNT_HISTORY[self.get_current_time] = self.Account

    def update_account_balance(self):
        self.ACCOUNT.buying_power = max(np.round(
            self.ACCOUNT.cash * self.LEVERAGE, 2), 0)

    def format_on_bar(self, bar, symbol: str):
        if self.DataFeed == 'yf':
            assert symbol, 'Symbol must be provided when using yf data feed - format_on_bar()'
            index = pd.MultiIndex.from_product(
                [[symbol], pd.to_datetime(bar.index, utc=True)], names=['symbol', 'date'])

            bar = pd.DataFrame(data={
                'open': np.array(bar['Open'].values).reshape(-1),
                'high': np.array(bar['High'].values).reshape(-1),
                'low': np.array(bar['Low'].values).reshape(-1),
                'close': np.array(bar['Close'].values).reshape(-1),
                'volume': np.array(bar['Volume'].values).reshape(-1),
            }, index=index, columns=['open', 'high', 'low', 'close', 'volume'])
            return bar
        else:
            print('DataFeed not supported')
            return None

    def format_on_quote(self, quote):
        data = IQuote(
            symbol=None,
            bid=quote.low,
            ask=quote.high,
            bid_size=0,
            ask_size=0,
            volume=quote.volume,
            timestamp=self.get_current_time
        )
        return data

    def format_on_trade_update(self, trade: ITradeUpdate):
        assert isinstance(
            trade, ITradeUpdate), 'Trade must be an instance of ITradeUpdate'
        return trade.order, trade.event

    def _applyTA(self, asset: IMarketDataStream):
        symbol = asset['symbol'] if asset.get(
            'feature') == None else asset['feature']
        if not asset.get('applyTA') or not asset.get('TA'):
            return
        print("Applying TA for: ", symbol)
        self.HISTORICAL_DATA[symbol]['bar'].ta.strategy(asset['TA'])

    def _load_historical_bar_data(self, asset: IMarketDataStream):
        try:
            bar_data_path = None
            symbol = asset['symbol'] if asset.get(
                'feature') == None else asset['feature']
            if asset.get('stored') and asset.get('stored_path'):
                bar_data_path = asset['stored_path'] + \
                    f'/bar/{asset['symbol']}_{asset["time_frame"]
                                              }_{self.START_DATE}-{self.END_DATE}.h5'

            if asset.get('stored'):
                if asset['stored_path']:
                    if os.path.exists(bar_data_path):
                        print("Loading data from ", bar_data_path)
                        self.HISTORICAL_DATA[symbol]['bar'] = pd.read_hdf(
                            bar_data_path)
                        # check if the data is empty
                        if self.HISTORICAL_DATA[symbol]['bar'].empty:
                            raise Exception({
                                "code": "no_data",
                                "data": {"symbol": symbol}
                            })
                        # apply the TA strategy
                        self._applyTA(asset)
                        print(
                            self.HISTORICAL_DATA[symbol]['bar'].describe())
                        return True
                    else:
                        raise BaseException({
                            "code": "file_not_found",
                            "data": {"path": bar_data_path}
                        })
                else:
                    raise BaseException({
                        "code": "path_not_provided",
                        "data": {"path": asset['stored_path']}
                    })
        except Exception as e:
            raise e
        except BaseException as e:
            print("Error: ", e.args[0]['code'], e.args[0]['data']['path'])

        if self.DataFeed == 'yf':
            # Populate the HISTORICAL_DATA with the bar data
            self.HISTORICAL_DATA[symbol]['bar'] = self.get_history(
                asset, self.START_DATE, self.END_DATE, asset['time_frame'], False)

            # check if the data is empty
            if self.HISTORICAL_DATA[symbol]['bar'].empty:
                raise Exception({
                    "code": "no_data",
                    "data": {"symbol": symbol}
                })
            # apply the TA strategy
            self._applyTA(asset)

            print("Loaded data for ", symbol)
            print(self.HISTORICAL_DATA[symbol]['bar'].describe())
            # if a stored path is provided save the data to the path
            if asset.get('stored_path'):
                # Create the directory if it does not exist
                Path(asset['stored_path']+'/bar').mkdir(
                    parents=True, exist_ok=True)
                # Save the data to the path
                print(self.HISTORICAL_DATA[symbol]['bar'].head(10))

                # Save the data to the path in hdf5 format
                print("Saving data to ", bar_data_path)
                self.HISTORICAL_DATA[symbol]['bar'].to_hdf(
                    bar_data_path, mode='a', key=asset["exchange"], index=True, format='table')

            return True

        else:
            print('DataFeed not supported')

    def streamMarketData(self, callback, assetStreams):
        """Listen to market data and call the callback function with the data"""
        super().streamMarketData(callback, assetStreams)
        loop = asyncio.new_event_loop()
        self.RUNNING_MARKET_STREAM = True
        TF = None  # strategy time frame
        hasFeature = False

        if self.MODE == IStrategyMode.BACKTEST:
            # Load Market data from yfinance for all assets
            self.HISTORICAL_DATA = {}

            for asset in tqdm(assetStreams, desc="Loading Historical Data"):
                symbol = asset['symbol'] if asset.get(
                    'feature') == None else asset['feature']
                if asset['type'] == 'bar':
                    self.HISTORICAL_DATA[symbol] = {}
                    # populate HISTORICAL_DATA
                    try:
                        if self._load_historical_bar_data(asset):
                            if self.MODE == IStrategyMode.BACKTEST:

                                if asset.get('feature') == None:
                                    # Create signals dataframe
                                    size = self.HISTORICAL_DATA[symbol
                                                                ]['bar'].shape[0]
                                    self.HISTORICAL_DATA[symbol]['signals'] = pd.DataFrame(data={"entries": np.zeros(size), "short_entries": np.zeros(size), "exits": np.zeros(
                                        size), "short_exits": np.zeros(size), "price": np.zeros(size), "qty": np.zeros(size)}, columns=["entries", "short_entries", "exits", "short_exits", "price", "qty"], index=self.HISTORICAL_DATA[asset['symbol']]['bar'].index)
                                    # .reindex_like(self.HISTORICAL_DATA[asset['symbol']]['bar'])
                                    self.HISTORICAL_DATA[symbol]['signals'][[
                                        'entries', 'exits', 'short_entries', 'short_exits']] = False
                                    self.HISTORICAL_DATA[symbol]['signals'][[
                                        'qty']] = 0
                                    self.HISTORICAL_DATA[symbol]['signals'][[
                                        'price']] = np.nan
                                    if not TF:
                                        # set the Main time frame to the first asset time frame
                                        TF = asset['time_frame']
                                else:
                                    # change the symbol to the feature symbol in the multi index
                                    self.HISTORICAL_DATA[symbol]['bar'].rename(
                                        index={asset['symbol']: symbol}, inplace=True)
                                    if hasFeature == False:
                                        # signal for the backtest to use dynamic time resoltion
                                        hasFeature = True

                    except Exception as e:
                        print("Removing Bar Stream",
                              asset['symbol'],  "\nError: ", e)
                        assetStreams.remove(asset)
                        continue

                else:
                    raise NotImplementedError(
                        f'Stream type not {self.DataFeed}supported')

            # Stream data to callback one by one for each IAsset

            while self.get_current_time <= self.END_DATE and self.RUNNING_MARKET_STREAM and len(assetStreams) > 0:
                try:
                    if self.VERBOSE > 0:
                        print("\nstreaming data for:",
                              self.get_current_time, "\n")
                    for asset in assetStreams:
                        isFeature = False
                        if asset['type'] == 'bar':
                            if asset.get('feature') == None:
                                symbol = asset['symbol']
                            else:
                                symbol = asset['feature']
                                isFeature = True

                            try:
                                if isFeature:
                                    if self.PreviousTime == None:
                                        barDatas = self.HISTORICAL_DATA[symbol]['bar'].loc[(self.HISTORICAL_DATA[symbol]['bar'].index.get_level_values('date') >= self.get_current_time.replace(
                                            tzinfo=datetime.timezone.utc)) & (self.HISTORICAL_DATA[symbol]['bar'].index.get_level_values('date') <= self.get_current_time.replace(tzinfo=datetime.timezone.utc))]
                                    else:
                                        barDatas = self.HISTORICAL_DATA[symbol]['bar'].loc[(self.HISTORICAL_DATA[symbol]['bar'].index.get_level_values('date') > self.PreviousTime.replace(
                                            tzinfo=datetime.timezone.utc)) & (self.HISTORICAL_DATA[symbol]['bar'].index.get_level_values('date') <= self.get_current_time.replace(tzinfo=datetime.timezone.utc))]
                                    if type(barDatas) == NoneType:
                                        continue
                                    elif barDatas.empty:
                                        continue
                                    else:
                                        for index in barDatas.index:
                                            loop.run_until_complete(
                                                callback(barDatas.loc[[index]], timeframe=asset['time_frame']))
                                else:
                                    # Get the current bar data with the index
                                    barData = self._get_current_bar(symbol)
                                    if type(barData) == NoneType:
                                        continue
                                    elif barData.empty:
                                        continue
                                    else:
                                        loop.run_until_complete(
                                            callback(barData, timeframe=asset['time_frame']))

                            except BaseException as e:
                                print("Error: ", e)
                                continue
                        else:
                            print('DataFeed not supported')

                    # Wait for all assets to be streamed and processed
                        # for future in as_completed(futures):
                        #     result = future.result()
                        #     print("Result: ", result)

                    self.BACKTEST_FlOW_CONTROL_BARRIER.wait()

                    # LOG the account history
                    self.update_account_history()
                    # Go to next time frame
                    # FIXME: Implement time frame increment
                    self.setCurrentTime(TF.get_next_time_increment(
                        self.get_current_time))
                    if self.get_current_time > self.END_DATE:
                        self.update_account_history()
                        self.BACKTEST_FlOW_CONTROL_BARRIER.abort()
                        break

                except BrokenBarrierError:
                    continue
                except Exception as e:
                    print("Error: ", e)
                    continue
            loop.run_until_complete(self.closeStream(assetStreams))
            loop.run_until_complete(self.closeTradeStream())
            self.BACKTEST_FlOW_CONTROL_BARRIER.abort()

            print("End of Market Stream")
        else:
            #  live data feed
            barStreamCount = 0
            for asset in assetStreams:
                self.HISTORICAL_DATA[asset['symbol']] = {}
                if asset['type'] == 'bar':
                    self.HISTORICAL_DATA[asset['symbol']
                                         ]['bar'] = pd.DataFrame()
                    barStreamCount += 1
                    # if asset.get('feature') == None:
                    #     TF = asset['time_frame']

            pool = ThreadPoolExecutor(max_workers=(
            barStreamCount), thread_name_prefix="MarketDataStream")
            for asset in assetStreams:
                if asset['type'] == 'bar':
                    # Stream data to callback one by one for each IAsset
                    async def PaperBarStreamer(asset: IMarketDataStream):
                        """
                        Stream data to the callback function
                        """
                        lastChecked = pd.Timestamp.now() - datetime.timedelta(minutes=self.FeedDelay) # last checked time for the asset stream data - FeedDelay
                        while self.RUNNING_MARKET_STREAM:
                            nextTimetoCheck = asset['time_frame'].get_next_time_increment(lastChecked)
                            await asyncio.sleep((nextTimetoCheck - lastChecked).total_seconds())

                            try:
                                # Get the current bar data with the index
                                barDatas = self._get_current_bar(
                                    asset["symbol"], asset['time_frame'], lastChecked)
                                if type(barDatas) == NoneType:
                                    continue
                                elif barDatas.empty:
                                    continue
                                else:
                                    # TODO: May need to check if this is a feature or not and get the data accordingly
                                    for idx in  range(0, len(barDatas)):
                                        bar = barDatas.iloc[[idx]]
                                        # if asset['time_frame'].is_time_increment(bar.index[0][1]) and (not (bar.index[0][1]) < lastChecked):
                                        if asset['time_frame'].is_time_increment(bar.index[0][1]) and (not (bar.index[0][1]) < lastChecked.replace(tzinfo=datetime.timezone.utc)):
                                            loop.run_until_complete(
                                                callback(bar, timeframe=asset['time_frame']))
    
                                        # loop.run_until_complete(
                                        #     callback(barData.loc[[index]], timeframe=asset['time_frame']))
                                    
                            except BaseException as e:
                                print("Error: ", e)
                                continue
                            
                            lastChecked = nextTimetoCheck
                    streamKey = f"{asset['symbol']}:{str(asset['time_frame'])}"
                    self._MARKET_STREAMS[streamKey] = loop.create_task(
                        PaperBarStreamer(asset), name=f"Market:{streamKey}")
            
                    # pool.submit(, asset)
            loop.run_forever()
                    


    async def closeStream(self,  assetStreams: List[IMarketDataStream]):
        if self.RUNNING_MARKET_STREAM:
            # _MARKET_STREAMS may nee to be part of the parant class
            for asset in assetStreams:
                streamKey = f"{asset['symbol']}:{str(asset['time_frame'])}"
                marketStream = self._MARKET_STREAMS.get(streamKey, None)
                if marketStream:
                    print("Closing Market Stream for: ", streamKey)
                    if type(marketStream) == asyncio.Task:
                        marketStream.cancel("Relenquishing Market Stream for: " +streamKey)
                    del self._MARKET_STREAMS[streamKey]
                    return True
                if len(self._MARKET_STREAMS) == 0:
                    self.RUNNING_MARKET_STREAM = False
                return True
        return False

    def close_position(self, symbol: str, qty=None, percent=None):
        position = self.get_position(symbol)
        if position:
            quantityToClose = 0
            counterPosistionSide = IOrderSide.BUY if position[
                'side'] == IOrderSide.SELL else IOrderSide.SELL
            if position['qty'] == 0:
                raise BaseException({
                    "code": "no_position",
                    "data": {"symbol": symbol}
                })
            if qty:
                if np.abs(qty) > np.abs(position['qty']):
                    raise BaseException({
                        "code": "invalid_qty",
                        "data": {"symbol": symbol, "qty": qty, "available": position['qty']}
                    })

                quantityToClose = qty

            elif percent:
                if percent > 100 or percent < 0:
                    raise BaseException({
                        "code": "invalid_percent",
                        "data": {"symbol": symbol, "percent": percent, "message": "Percent must be between 0 and 100"}
                    })
                quantityToClose = np.abs(position['qty']) * (percent / 100)

            marketCloseOrder = IOrder(
                order_id=uuid.uuid4(),
                asset=self.get_ticker_info(symbol),
                limit_price=None,
                filled_price=None,
                stop_price=None,
                qty=quantityToClose,
                filled_qty=None,
                side=counterPosistionSide,
                type=IOrderType.MARKET,
                time_in_force=ITimeInForce.GTC,
                status=ITradeUpdateEvent.NEW,
                order_class=IOrderClass.SIMPLE,
                created_at=self.get_current_time,
                updated_at=self.get_current_time,
                submitted_at=self.get_current_time,
                filled_at=None,
                legs=None
            )
            """FIXME: This should also remove the active orders that need to be closed or reduced in size
                right now it only closes the position and does not remove the active order from the active orders list.
                This would mean that the order would still be active and could be filled later on, thus adjusting the position size.
                which would make the results inconsistent with the actual orders placed.

                self.CLOSE_ORDERS.append(marketCloseOrder)
                then we can process the rest in the self._update_order() method
            """
            self._update_order(marketCloseOrder)
            return marketCloseOrder
        else:
            raise BaseException({
                "code": "no_position",
                "data": {"symbol": symbol}
            })

    def close_all_positions(self):
        activePositions = self.get_positions()
        for symbol in activePositions:
            if activePositions.get(symbol) != None:
                self.close_position(symbol, qty=activePositions[symbol]['qty'])
        return True

    @property
    def get_current_time(self) -> datetime.datetime:
        if self.MODE == IStrategyMode.BACKTEST:
            return self.CurrentTime
        else:
            return datetime.datetime.now() - datetime.timedelta(minutes=self.FeedDelay)

    def setCurrentTime(self, time: datetime.datetime):
        self.PreviousTime = self.CurrentTime
        self.CurrentTime = time

    def _get_current_bar(self, symbol: str, timeFrame: ITimeFrame = ITimeFrame(unit=ITimeFrameUnit.Minute, amount=1), lastChecked: datetime.datetime = None) -> pd.DataFrame:
        currentBar = None

        def find_next_bar(start: datetime.datetime, end: datetime.datetime):
            idx = pd.IndexSlice
            nonlocal currentBar
            nonlocal symbol
            try:
                if self.HISTORICAL_DATA[symbol]['bar'].empty:
                    return None
                # Try to get the exact current time bar
                currentBar = self.HISTORICAL_DATA[symbol]['bar'].loc[idx[symbol, end:end], :]
                # print("Current Bar: ", currentBar)
                if currentBar.empty:
                    currentBar = self.HISTORICAL_DATA[symbol]['bar'].loc[idx[symbol, start:end], :]
                    if currentBar.empty:
                        return None
                return True
            except KeyError:
                return None

        def convert_to_utc(time: datetime.datetime):
            # time.astimezone(datetime.timezone.utc) # Some brokers may need this line as right now your just falsely assuming the time is in UTC when it may not be
            return time.replace(tzinfo=datetime.timezone.utc)

        if self.MODE == IStrategyMode.BACKTEST:
            if symbol in self.HISTORICAL_DATA:

                current_time = convert_to_utc(self.get_current_time)
                if self.PreviousTime == None:
                    previous_time = current_time
                else:
                    previous_time = convert_to_utc(self.PreviousTime)

                found = find_next_bar(previous_time, current_time)
                if found == None:
                    return None

            else:
                raise BaseException({
                    "code": "symbol_not_found",
                    "data": {"symbol": symbol}
                })
        else:
            # live data feed
            assert timeFrame, 'TimeFrame must be provided when using live data feed - _get_current_bar()'
            useTime = self.get_current_time
            
            
            tf_current_time = timeFrame.get_time_increment(useTime)

            current_time = convert_to_utc(tf_current_time)
            # if self.PreviousTime == None:
            #     # Default to the current time for backtesting 
            #     previous_time = current_time
            # else:
            #     previous_time = convert_to_utc(self.PreviousTime)
            # TODO: Else statement needs to be tested as we should just get the latest bar from the previous time
            previous_time = convert_to_utc(lastChecked) if lastChecked else current_time

            find_next_bar(previous_time, current_time)
            if type(currentBar) == NoneType or currentBar.empty:
                try:
                    if not self.TICKER_INFO[symbol]:
                        self.get_ticker_info(symbol)
                    getBarsFrom = timeFrame.add_time_increment(tf_current_time, -2)
                    # TODO: Could be using previous_time as we now track the last checked time for each asset stream
                    recentBars = self.get_history(
                        self.TICKER_INFO[symbol], getBarsFrom, timeFrame.get_next_time_increment(useTime), timeFrame)
                    self.HISTORICAL_DATA[symbol]['bar'] = pd.concat(
                        [self.HISTORICAL_DATA[symbol]['bar'], recentBars])
                    # Remove duplicates keys in the history as sometimes if we get duplicates
                    self.HISTORICAL_DATA[symbol]['bar'] = self.HISTORICAL_DATA[symbol]['bar'].loc[~self.HISTORICAL_DATA[symbol]['bar'].index.duplicated(
                        keep='first')]

                    found = find_next_bar(previous_time, current_time)
                    if found == None:
                        # Default to the last bar
                        currentBar = self.HISTORICAL_DATA[symbol]['bar'].iloc[-1, :]
                        return None
                except KeyError:
                    return None
            # elif currentBar.empty:
            #     return None

        return currentBar

        # raise NotImplementedError(f'Mode {self.MODE} not supported')

    def get_VBT_results(self, timeFrame: ITimeFrame) -> dict[str, vbt.Portfolio]:
        """returns the backtest results from Vector BT - profit and loss,  profit and loss percentage, number of orders executed and filled, cag sharp rattio, percent win. etc."""
        results: dict[str, vbt.Portfolio] = {}

        def calc_expectancy_ratio(trades: vbt.EntryTrades):
            if len(trades) == 0:
                return np.nan

            returns = trades.returns.values
            winners = returns[returns > 0]
            losers = returns[returns <= 0]

            pct_winners = len(winners) / len(returns)
            pct_losers = len(losers) / len(returns)

            avg_win = np.mean(winners) if len(winners) > 0 else 0
            avg_loss = abs(np.mean(losers)) if len(losers) > 0 else 1
            R = avg_win / avg_loss

            E = pct_winners * R - pct_losers
            return E

        expectancy_ratio = (
            "Expectancy Ratio",
            dict(
                title="Expectancy Ratio",
                calc_func=calc_expectancy_ratio,
                resolve_trades=True
            )
        )

        if self.MODE == IStrategyMode.BACKTEST:
            for asset in tqdm(self.HISTORICAL_DATA.keys(), desc="Running VBT Backtest"):
                if isinstance(self.HISTORICAL_DATA[asset].get('signals'), pd.DataFrame):
                    try:
                        vbtParams = {
                            "init_cash": self.STARTING_CASH,
                            "open": self.HISTORICAL_DATA[asset]['bar']['open'].reset_index(level='symbol', drop=True),
                            "high": self.HISTORICAL_DATA[asset]['bar']['high'].reset_index(level='symbol', drop=True),
                            "low": self.HISTORICAL_DATA[asset]['bar']['low'].reset_index(level='symbol', drop=True),
                            "close": self.HISTORICAL_DATA[asset]['bar']['close'].reset_index(level='symbol', drop=True),
                            "entries": self.HISTORICAL_DATA[asset]['signals']['entries'].reset_index(level='symbol', drop=True),
                            "short_entries": self.HISTORICAL_DATA[asset]['signals']['short_entries'].reset_index(level='symbol', drop=True),
                            "exits": self.HISTORICAL_DATA[asset]['signals']['exits'].reset_index(level='symbol', drop=True),
                            "short_exits": self.HISTORICAL_DATA[asset]['signals']['short_exits'].reset_index(level='symbol', drop=True),
                            "price": self.HISTORICAL_DATA[asset]['signals']['price'].reset_index(level='symbol', drop=True),
                            "size":  self.HISTORICAL_DATA[asset]['signals']['qty'].abs().reset_index(level='symbol', drop=True),
                            "freq": timeFrame.unit.value[0].lower(),
                            "accumulate": True
                        }

                        # run the backtest
                        print("Running backtest with Vector BT for:", asset)
                        results[asset] = vbt.Portfolio.from_signals(
                            **vbtParams)
                        # results[asset] = vbt.Portfolio.from_orders(
                        #     **vbtParams)
                        print(results[asset].stats(
                            metrics=[*vbt.Portfolio.metrics, expectancy_ratio]))
                    except Exception as e:
                        print("Failed to run backtest for ", asset, e)
                        continue
                else:
                    print("No signals for feature ", asset)
                    continue
        else:
            raise NotImplementedError(f'Mode {self.MODE} not supported')
        return results

    @property
    def Account(self) -> IAccount:
        """ Returns the state of the strategy."""
        self.update_account_balance()
        return self.ACCOUNT

    @Account.setter
    def Account(self, account: IAccount):
        """ Sets the state of the strategy."""
        cash = account.cash
        if cash and cash != self.ACCOUNT.cash:
            self.ACCOUNT.cash = max(cash, 0)
        equity = account.get('equity')
        if equity and equity != self.ACCOUNT.equity:
            self.ACCOUNT.equity = max(equity, 0)

        if account.get('leverage') and account.leverage != self.ACCOUNT.leverage:
            self.ACCOUNT.leverage = account.leverage

        self.update_account_balance()


if __name__ == '__main__':
    # os.path.join(os.path.dirname(__file__), 'data')
    broker = PaperBroker(cash=1_000_000, start_date=datetime(
        2024, 5, 27), end_date=datetime(2024, 5, 31))
